"""Repulsion Attraction Neutralization WEFE implementation."""

from collections.abc import Sequence
from copy import deepcopy
from typing import Any

import numpy as np
from sklearn.decomposition import PCA
from tqdm import tqdm

from wefe.debias.base_debias import BaseDebias
from wefe.preprocessing import EmbeddingDict, get_embeddings_from_tuples
from wefe.utils import check_is_fitted
from wefe.word_embedding_model import WordEmbeddingModel

try:
    import torch
    from torch import nn
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "PyTorch is required to run RepulsionAttractionNeutralization method. "
        "Visit https://pytorch.org/ to install it."
    ) from e


class RAN_Loss(nn.Module):
    """Loss for Repulsion Attraction Neutralization method."""

    def __init__(
        self,
        w_b: np.ndarray,
        w: np.ndarray,
        repulsion_set: list[np.ndarray],
        bias_direction: np.ndarray,
    ) -> None:
        """Initialize a RAN_Loss instance.

        Parameters
        ----------
            w_b: np.array
                Debiased embedding of word
            w: np.array
                Original embedding of word
            repulsion_set: List[np.ndarray]
                Set of embeddings to be repeled from word
            bias_direction: np.array

        """
        super().__init__()
        self.w = w
        self.w_b = nn.Parameter(w_b)
        self.repulsion_set = repulsion_set
        self.bias_direction = bias_direction

    def _repulsion(self) -> torch.Tensor:
        if not isinstance(self.repulsion_set, bool):
            cos_similarity = self._cosine_similarity(self.w_b, self.repulsion_set)
            repulsion = torch.abs(cos_similarity).mean(dim=0)
        else:
            repulsion = 0
        return repulsion

    def _attraction(self) -> torch.Tensor:
        attraction = (
            torch.abs(torch.cosine_similarity(self.w_b[None, :], self.w[None, :]) - 1)[
                0
            ]
            / 2
        )
        return attraction

    def _neutralization(self) -> torch.Tensor:
        neutralization = torch.abs(self.w_b.dot(self.bias_direction)).mean(dim=0)
        return neutralization

    def _objective_function(
        self,
        w1: np.number,
        w2: np.number,
        w3: np.number,
    ) -> torch.Tensor:
        return (
            self._repulsion() * w1
            + self._attraction() * w2
            + self._neutralization() * w3
        )

    def _cosine_similarity(
        self, w: torch.Tensor, set_vectors: torch.Tensor
    ) -> torch.Tensor:
        return torch.matmul(set_vectors, w) / (set_vectors.norm(dim=1) * w.norm(dim=0))


class RAN(nn.Module):
    """NN that performs the optimization by gradient descent of the loss function."""

    def __init__(
        self,
        model: WordEmbeddingModel,
        word: str,
        w_b: np.ndarray,
        w: np.ndarray,
        repulsion_set: list[np.ndarray],
        bias_direction: np.ndarray,
        weights: Sequence = [0.33, 0.33, 0.33],
    ) -> None:
        """Initialize a RAN instance.

        Parameters
        ----------
            model: WordEmbeddingModel
                The word embedding model to debias.
            word: str
                Word to be debiased
            w_b: np.array
                Debiased embedding of word
            w: np.array
                Original embedding of word
            repulsion_set: List[np.ndarray]
                Set of embeddings to be repeled from word
            bias_direction: np.array
            weights: Sequence, optional
                weights λi that determine the relative importance of one
                objective function (repulsion, attraction, neutralization)
                over another. by Defaults [0.33, 0.33, 0.33].

        """
        super().__init__()

        self.model = model
        self.word = word
        self.w = torch.FloatTensor(np.array(w)).requires_grad_(True)

        if len(repulsion_set) == 0:
            self.repulsion_set = False
        else:
            self.repulsion_set = torch.FloatTensor(
                np.array(repulsion_set)
            ).requires_grad_(True)

        self.w_b = nn.Parameter(w_b)

        self.bias_direction = torch.FloatTensor(np.array(bias_direction))

        self.weights = weights

        self.loss = RAN_Loss(self.w, self.w_b, self.repulsion_set, self.bias_direction)

    def _forward(self) -> torch.Tensor:
        w1, w2, w3 = self.weights
        return self.loss._objective_function(w1, w2, w3)


class RepulsionAttractionNeutralization(BaseDebias):
    r"""Repulsion Attraction Neutralization method.

    .. warning::

        This method only works if Pytorch is installed. If you do not have it
        installed, check
        `<https://pytorch.org/get-started/locally/>`_ for further information.

    This method allow reducing the bias of an embedding model creating a transformation
    such that the stereotypical information is minimized with minimal semantic offset.
    This transformation bases its operations on:

    1. Repelling embeddings from neighbours with a high value of indirect bias
       (indicating a strong association due to bias), to minimize the
       bias based illicit associations.
    2. Attracting debiased embeddings to the original representation, to minimize the
       loss of semantic meaning.
    3. Neutralizing the bias direction of each word, minimizing its bias to any
       particular group.

    This method is binary because it only allows 2 classes of the same bias
    criterion, such as male or female.

    .. note::

        For a multiclass debias (such as for Latinos, Asians and Whites), it is
        recommended to visit
        :class:`~wefe.debias.multiclass_hard_debias.MulticlassHardDebias` class.

    The steps followed to perform the debias are:

    1. Identify a bias subspace through the defining sets. In the case of
       gender, these could be e.g. ``[['woman', 'man'], ['she', 'he'], ...]``

    2. A multi-objective optimization is performed. For each vector :math:`w`  in the
       target set it is found its debias counterpart :math:`w_d`  by solving:

    .. math::

        argmin(F_r(w_d),F_a(w_d),F_n(w_d))

    where :math:`Fr`, :math:`Fa`, :math:`Fn` are repulsion, attraction and
    neutralization functions defined as the following:

    .. math::

        F_r(w_d) =  \sum |cos(w_d,n_i)| / |S|

    .. math::

        F_a(w_d) = |cos(w_d,w)-1|/2

    .. math::

        F_n(w_d) = |cos(w_d,g)|

    The optimization is performed by formulating a single objective:

    .. math::

        F(w_d) =  \lambda_1 F_r(w_d) + \lambda_2 F_a(w_d) + \lambda_3 F_n(w_d)

    In the original implementation is define a preserve set :math:`(V_p)` corresponding
    to words for which gender carries semantic importance, this words are not
    included in the debias process.

    In WEFE this words would be the ones included in the ignore parameter of the
    transform method. The words that are not present in :math:`V_p` are the ones to be
    included in the debias process and form part of the debias set :math:`(V_d)`,
    in WEFE this words can be specified in the target parameter of the transform method.

    Examples
    --------
    The following example shows how to execute Repulsion Attraction
    Neutralization method that reduces bias in a word embedding model:

    >>> from wefe.debias.repulsion_attraction_neutralization import (
    ...   RepulsionAttractionNeutralization
    ... )
    >>> from wefe.utils import load_test_model
    >>> from wefe.datasets import fetch_debiaswe
    >>>
    >>> # load the model (in this case, the test model included in wefe)
    >>> model = load_test_model()
    >>> # load definitional pairs, in this case definitinal pairs included in wefe
    >>> debiaswe_wordsets = fetch_debiaswe()
    >>> definitional_pairs = debiaswe_wordsets["definitional_pairs"]
    >>>
    >>> # instance and fit the method
    >>> ran = RepulsionAttractionNeutralization().fit(
    ...     model = model,
    ...     definitional_pairs= definitional_pairs
    ...   )
    >>> # execute the debias passing words over a set of target words
    >>> debiased_model = ran.transform(
    ...    model = model, target = ['doctor','nurse','programmer']
    ... )
    Copy argument is True. Transform will attempt to create a copyof the original model.
    This may fail due to lack of memory.
    Model copy created successfully.
    >>> # if you don't want a set of words to be debiased include them in the ignore set
    >>> gender_specific = debiaswe_wordsets["gender_specific"]
    >>> debiased_model = ran.transform(
    ...    model = model, ignore= gender_specific
    ... ) # doctest: +SKIP


    References
    ----------
    | [1]: Kumar, Vaibhav, Tenzin Singhay Bhotia y Tanmoy Chakraborty: Nurse
      is Closer to Woman than Surgeon? Mitigating Gender-Biased Proximities
      in Word Embeddings. CoRR,abs/2006.01938, 2020.
    | https://arxiv.org/abs/2006.01938
    | [2]: https://github.com/TimeTraveller-San/RAN-Debias

    """

    name = "Repulsion attraction Neutralization"
    short_name = "RAN"

    def __init__(
        self,
        pca_args: dict[str, Any] = {"n_components": 10},
        verbose: bool = False,
        criterion_name: str | None = None,
        epochs: int = 300,
        theta: float = 0.05,
        n_neighbours: int = 100,
        learning_rate: float = 0.01,
        weights: list[float] = [0.33, 0.33, 0.33],
    ) -> None:
        """Initialize a Repulsion Attraction Neutralization Debias instance.

        Parameters
        ----------
        pca_args : Dict[str, Any], optional
            Arguments for the PCA that is calculated internally in the
            identification of the bias subspace,
            by default {"n_components": 10}
        verbose : bool, optional
            True will print informative messages about the debiasing process,
            by default False.
        criterion_name : Optional[str], optional
            The name of the criterion for which the debias is being executed,
            e.g., 'Gender'. This will indicate the name of the model returning
            transform, by default None
        epochs : int, optional
            number of times that the minimization is done. By default 300
        theta: float, optional
            Indirect bias threshold to select neighbours for the repulsion set.
            By default 0.05
        n_neighbours: int, optional
            Number of neighbours to be consider for the repulsion set.
            By default 100
        learning_rate: float, optional
            Learning rate to be used by the optimizer during the optimization.
            By default 0.01
        weights: List[float], optional
            List of the 3 initial weights to be used. By default [0.33,0.33,0.33]

        """
        # check verbose
        if not isinstance(verbose, bool):
            raise TypeError(f"verbose should be a bool, got {verbose}.")
        self.pca_args = pca_args
        self.verbose = verbose

        if criterion_name is None or isinstance(criterion_name, str):
            self.criterion_name_ = criterion_name
        else:
            raise ValueError(f"criterion_name should be str, got: {criterion_name}")

        if not isinstance(epochs, int):
            raise TypeError(f"epochs should be a int, got {epochs}.")
        self.epochs = epochs

        if not isinstance(theta, float):
            raise TypeError(f"theta should be a float, got {theta}.")
        self.theta = theta

        if not isinstance(n_neighbours, int):
            raise TypeError(f"n_neighbours should be a int, got {n_neighbours}.")
        self.n_neighbours = n_neighbours

        if not isinstance(weights, list):
            raise TypeError(f"weights should be a list, got {weights}.")
        self.weights = weights

        if not isinstance(learning_rate, float):
            raise TypeError(f"learning_rate should be a float, got {learning_rate}.")
        self.learning_rate = learning_rate

    def _identify_bias_subspace(
        self,
        defining_pairs_embeddings: list[EmbeddingDict],
        verbose: bool = False,
    ) -> PCA:
        matrix = []
        for embedding_dict_pair in defining_pairs_embeddings:
            # Get the center of the current defining pair.
            pair_embeddings = np.array(list(embedding_dict_pair.values()))
            center = np.mean(pair_embeddings, axis=0)
            # For each word, embedding in the defining pair:
            for embedding in embedding_dict_pair.values():
                # Subtract the center of the pair to the embedding
                matrix.append(embedding - center)
        matrix = np.array(matrix)  # type: ignore

        pca = PCA(**self.pca_args)
        pca.fit(matrix)

        if verbose:
            explained_variance = pca.explained_variance_ratio_
            print(
                f"PCA variance explained: {explained_variance[0 : pca.n_components_]}"
            )

        return pca

    def _check_sets_size(
        self,
        sets: Sequence[Sequence[str]],
        set_name: str,
    ) -> None:
        for idx, set_ in enumerate(sets):
            if len(set_) != 2:
                adverb = "less" if len(set_) < 2 else "more"

                raise ValueError(
                    f"The {set_name} pair at position {idx} ({set_}) has {adverb} "
                    f"words than allowed by {self.name}: "
                    f"got {len(set_)} words, expected 2."
                )

    def _indirect_bias(
        self, w: np.ndarray, v: np.ndarray, bias_direction: np.ndarray
    ) -> float:
        wv = np.dot(w, v)
        w_orth = w - np.dot(w, bias_direction) * bias_direction
        v_orth = v - np.dot(v, bias_direction) * bias_direction
        cos_wv_orth = np.dot(w_orth, v_orth) / (
            np.linalg.norm(w_orth) * np.linalg.norm(v_orth)
        )
        bias = (wv - cos_wv_orth) / wv
        return bias

    def _get_neighbours(
        self, model: WordEmbeddingModel, word: str, n_neighbours: int
    ) -> list[str]:
        similar_words = model.wv.most_similar(positive=word, topn=n_neighbours)
        similar_words = list(list(zip(*similar_words, strict=False))[0])
        return similar_words

    def _get_repulsion_set(
        self,
        model: WordEmbeddingModel,
        word: str,
        bias_direction: np.ndarray,
        theta: float,
        n_neighbours: int,
    ) -> list[np.ndarray]:
        r"""Obtain the embeddings of the words that should be repealed from word arg.

        These words are the n_neighbours more similar to "word" whose indirect bias is
        greater than theta.

        Parameters
        ----------
            model: WordEmbeddingModel
                The word embedding model to debias.
            word: str
                Word to debias.
            bias_direction: np.ndarray
                The bias subspace
            theta: float
                Threshold to include neighbours in repulsion set.
            n_neighbours: int
                Number of neighbours to be considered.

        Returns
        -------
        List[np.ndarray]
            The list of embeddings that compose the repulsion set.

        """
        neighbours = self._get_neighbours(model, word, n_neighbours)
        repulsion_set = []
        for neighbour in neighbours:
            if (
                self._indirect_bias(model[neighbour], model[word], bias_direction)
                > theta
            ):
                repulsion_set.append(model[neighbour])
        return repulsion_set

    def _debias(
        self,
        model: WordEmbeddingModel,
        word: str,
        w: np.ndarray,
        w_b: np.ndarray,
        bias_direction: np.ndarray,
        repulsion_set: list[np.ndarray],
        learning_rate: float,
        epochs: int,
        weights: list[float],
    ) -> torch.Tensor:
        ran = RAN(
            model,
            word,
            w_b,
            w,
            repulsion_set,
            bias_direction,
            weights,
        )
        optimizer = torch.optim.Adam(ran.parameters(), lr=learning_rate)
        for _ in range(epochs):
            optimizer.zero_grad()
            out = ran._forward()
            out.backward()
            optimizer.step()
        debiased_vector = ran.w_b
        return debiased_vector / torch.norm(debiased_vector)

    def _init_vector(self, model: WordEmbeddingModel, word: str) -> torch.Tensor:
        v = deepcopy(model[word])
        return torch.FloatTensor(np.array(v))

    def fit(
        self,
        model: WordEmbeddingModel,
        definitional_pairs: Sequence[Sequence[str]],
    ) -> BaseDebias:
        """Compute the bias direction.

        Parameters
        ----------
        model : WordEmbeddingModel
            The word embedding model to debias.
        definitional_pairs : Sequence[Sequence[str]]
            A sequence of string pairs that will be used to define the bias
            direction. For example, for the case of gender debias, this list
            could be [['woman', 'man'], ['girl', 'boy'], ['she', 'he'],
            ['mother', 'father'], ...].

        Returns
        -------
        BaseDebias
            The debias method fitted.

        """
        # Check arguments types
        self._check_sets_size(definitional_pairs, "definitional")
        self.definitional_pairs_ = definitional_pairs

        # ------------------------------------------------------------------------------
        # Obtain the embedding of each definitional pairs.
        if self.verbose:
            print("Obtaining definitional pairs.")

        self.definitional_pairs_embeddings_ = get_embeddings_from_tuples(
            model=model,
            sets=definitional_pairs,
            sets_name="definitional",
            warn_lost_sets=self.verbose,
            normalize=True,
            verbose=self.verbose,
        )

        # -----------------------------------------------------------------------------:
        # Identify the bias subspace using the defining pairs.
        if self.verbose:
            print("Identifying the bias subspace.")

        self.pca_ = self._identify_bias_subspace(
            self.definitional_pairs_embeddings_,
            self.verbose,
        )
        self.bias_direction_ = self.pca_.components_[0]
        return self

    def transform(
        self,
        model: WordEmbeddingModel,
        target: list[str] | None = None,
        ignore: list[str] | None = [],
        copy: bool = True,
    ) -> WordEmbeddingModel:
        """Execute Repulsion Attraction Neutralization Debias over the provided model.

        Parameters
        ----------
        model : WordEmbeddingModel
            The word embedding model to debias.
        target : Optional[List[str]], optional
            If a set of words is specified in target, the debias method will
            be performed only on the word embeddings of this set. If `None`
            is provided, the debias will be performed on all words (except
            those specified in ignore). by default `None`.
        ignore : Optional[List[str]], optional
            If target is `None` and a set of words is specified in ignore,
            the debias method will perform the debias in all words except
            those specified in this set, by default `None`.
        copy : bool, optional
            If `True`, the debias will be performed on a copy of the model.
            If `False`, the debias will be applied on the same model delivered,
            causing its vectors to mutate.
            **WARNING:** Setting copy with `True` requires RAM at least 2x of
            the size of the model, otherwise the execution of the debias may
            raise to `MemoryError`, by default True.
        WordEmbeddingModel
            The debiased embedding model.

        """
        # check if the following attributes exist in the object.
        check_is_fitted(
            self,
            [
                "bias_direction_",
                "pca_",
                "definitional_pairs_embeddings_",
                "definitional_pairs_",
            ],
        )

        # ------------------------------------------------------------------------------
        # Copy
        if copy:
            print(
                "Copy argument is True. Transform will attempt to create a copy"
                "of the original model. This may fail due to lack of memory."
            )
            model = deepcopy(model)
            print("Model copy created successfully.")

        else:
            print(
                "copy argument is False. The execution of this method will mutate"
                "the original model."
            )

        # Normalize embeddings
        if self.verbose:
            print("Normalizing embeddings.")
        model.normalize()

        if self.verbose:
            print(
                f"Executing Repulsion attraction Neutralization Debias on {model.name}"
            )

        # If none target words are provided the debias process is executed
        # over the entire vocabulary
        if not target:
            target = list(model.vocab.keys())

        debiased = {}
        for word in tqdm(target):
            if word in ignore or word not in model:
                continue
            w = model[word]
            w_b = self._init_vector(model, word)
            repulsion = self._get_repulsion_set(
                model, word, self.bias_direction_, self.theta, self.n_neighbours
            )
            new_embedding = self._debias(
                model,
                word,
                w,
                w_b,
                self.bias_direction_,
                repulsion,
                self.learning_rate,
                self.epochs,
                self.weights,
            )
            debiased[word] = new_embedding.detach().numpy()

        if self.verbose:
            print("Updating debiased vectors")

        for word in tqdm(debiased):
            model.update(word, debiased[word].astype(model.wv.vectors.dtype))

        # ------------------------------------------------------------------------------
        # # Generate the new KeyedVectors
        if self.criterion_name_ is None:
            new_model_name = f"{model.name}_debiased"
        else:
            new_model_name = f"{model.name}_{self.criterion_name_}_debiased"
        model.name = new_model_name

        if self.verbose:
            print("Done!")

        return model



"""Double Hard Debias WEFE implementation."""

from copy import deepcopy
import operator
from typing import Any

import numpy as np
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA, IncrementalPCA
from sklearn.metrics import pairwise_distances
from tqdm import tqdm

from wefe.debias.base_debias import BaseDebias
from wefe.preprocessing import get_embeddings_from_tuples
from wefe.utils import check_is_fitted
from wefe.word_embedding_model import WordEmbeddingModel


class DoubleHardDebias(BaseDebias):
    """Double Hard Debias Method.

    This method allow reducing the bias of an embedding model through geometric
    operations between embeddings.
    This method is binary because it only allows 2 classes of the same bias
    criterion, such as male or female.

    The main idea of this method is:

    1. Identify a bias subspace through the defining sets. In the case
    of gender, these could be e.g. ``[['woman', 'man'], ['she', 'he'], ...]``

    2. Find the dominant directions of the entire set of vectors by doing a
    Principal components analysis over it.

    3. Get the target words by finding the most biased words, this is
    the words tha are closest to the representation of each bias group. In
    case of gender ``'he'`` and ``'she'``.

    3. Try removing each component resulting of PCA and remove also the bias
    direction to every vector in the target set and find which component
    reduces bias the most.

    4. Remove the dominant direction that most reduces bias and remove also
    the bias direction of the vectors in the target set.

    Examples
    --------
    The following example shows how to execute Double Hard Debias method
    that reduces bias in a word embedding model:

    >>> from wefe.debias.double_hard_debias import DoubleHardDebias
    >>> from wefe.utils import load_test_model
    >>> from wefe.datasets import fetch_debiaswe
    >>>
    >>> # load the model (in this case, the test model included in wefe)
    >>> model = load_test_model()
    >>> # load definitional pairs, in this case definitinal pairs included in wefe
    >>> debiaswe_wordsets = fetch_debiaswe()
    >>> definitional_pairs = debiaswe_wordsets["definitional_pairs"]
    >>>
    >>> # instance and fit the method including bias representation words,
    >>> # in case of gender definitional_pairs=[['he', 'she'], ...]
    >>> dhd = DoubleHardDebias().fit(
    ...     model=model,
    ...     definitional_pairs=definitional_pairs,
    ...     bias_representation=['he','she'])
    >>> # execute the debias, if you don't want a set of words to be debiased
    >>> # include them in the ignore set.
    >>> gender_specific = debiaswe_wordsets["gender_specific"]
    >>>
    >>> debiased_model = dhd.transform(
    ...     model=model, ignore=gender_specific
    ... )
    Copy argument is True. Transform will attempt to create a copy of the original
    model. This may fail due to lack of memory.
    Model copy created successfully.

    If you want the debiased to be performed over a specific set of words you can
    specify them in the target parameter

    >>> debiased_model = dhd.transform(
    ... model=model, target = ['doctor','nurse','programmer','teacher']
    ... )
    Copy argument is True. Transform will attempt to create a copy of the original
    model. This may fail due to lack of memory.
    Model copy created successfully.

    References
    ----------
    | [1]: Wang, Tianlu, Xi Victoria Lin, Nazneen Fatema Rajani, Bryan McCann,
           Vicente Or-donez y Caiming Xiong.
    |      Double-Hard Debias: Tailoring Word Embeddings for GenderBias Mitigation.
    |      CoRR, abs/2005.00965, 2020.https://arxiv.org/abs/2005.00965.
    | [2]: https://github.com/uvavision/Double-Hard-Debias

    """

    name = "Double Hard Debias"
    short_name = "DHD"

    def __init__(
        self,
        pca_args: dict[str, Any] = {"n_components": 10},
        verbose: bool = False,
        criterion_name: str | None = None,
        incremental_pca: bool = True,
        n_words: int = 1000,
        n_components: int = 4,
    ) -> None:
        """Initialize a Double Hard Debias instance.

        Parameters
        ----------
        pca_args : Dict[str, Any], optional
            Arguments for the PCA that is calculated internally in the identification
            of the bias subspace, by default {"n_components": 10}
        verbose : bool, optional
            True will print informative messages about the debiasing process,
            by default False.
        criterion_name : Optional[str], optional
            The name of the criterion for which the debias is being executed,
            e.g., 'Gender'. This will indicate the name of the model
            returning transform, by default None
        incremental_pca: bool, optional
            If `True`, incremental pca will be used over the entire set of
            vectors.
            If `False`, pca will be used over the entire set of vectors.
            **WARNING:** Running pca over the entire set of vectors may
            raise to `MemoryError`,  by default True.
        n_words: int, optional
            Number of target words to be used for each bias group.
            By default 1000
        n_components: int, optional
            Numbers of components of PCA to be used to explore the one that
            reduces bias the most. Usually the best one is close to embedding
            dimension/100. By default 4.

        """
        # check verbose
        if not isinstance(verbose, bool):
            raise TypeError(f"verbose should be a bool, got {verbose}.")

        # check incremental pca
        if not isinstance(incremental_pca, bool):
            raise TypeError(f"incremental_pca should be a bool, got {verbose}.")

        if not isinstance(pca_args, dict):
            raise TypeError(f"pca_args should be a dict, got {verbose}.")

        self.pca_args = pca_args
        self.verbose = verbose

        if incremental_pca:
            self.pca_type = IncrementalPCA(**self.pca_args)
        else:
            self.pca_type = PCA(svd_solver="randomized", **self.pca_args)

        if criterion_name is None or isinstance(criterion_name, str):
            self.criterion_name_ = criterion_name
        else:
            raise ValueError(f"criterion_name should be str, got: {criterion_name}")

        if not isinstance(n_words, int):
            raise TypeError(f"n_words should be int, got: {n_words}")
        self.n_words = n_words

        if not isinstance(n_components, int):
            raise TypeError(f"n_components should be int, got: {n_components}")
        self.n_components = n_components

    def _check_sets_size(
        self,
        sets: list[list[str]],
        set_name: str,
    ) -> None:
        for idx, set_ in enumerate(sets):
            if len(set_) != 2:
                adverb = "less" if len(set_) < 2 else "more"

                raise ValueError(
                    f"The {set_name} pair at position {idx} ({set_}) has {adverb} "
                    f"words than allowed by {self.name}: "
                    f"got {len(set_)} words, expected 2."
                )

    def _similarity(self, u: list[np.ndarray], v: list[np.ndarray]) -> np.ndarray:
        return 1 - pairwise_distances(u, v, metric="cosine")

    def _bias_by_projection(
        self,
        model: WordEmbeddingModel,
        target: list[str],
        ignore: list[str],
        bias_representation: list[str],
    ) -> dict[str, float]:
        word1 = model[bias_representation[0]]
        word2 = model[bias_representation[1]]
        if target:
            vectors = [model[word] for word in target if word in model]
            words = target
        else:
            vectors = model.wv.vectors
            words = list(model.vocab.keys())
        vectors = model.wv.vectors
        similarities_vectors = (
            self._similarity([word1], vectors)[0]
            - self._similarity([word2], vectors)[0]
        )

        words = list(model.vocab.keys())
        similarities = dict(zip(words, similarities_vectors, strict=False))
        for word in ignore:
            if word in similarities:
                similarities.pop(word)
        return similarities

    def get_target_words(
        self,
        model: WordEmbeddingModel,
        target: list[str],
        ignore: list[str],
        n_words: int,
        bias_representation: list[str],
    ) -> list[str]:
        """Obtain target words to be debiased.

        This is done by searching the "n_words" most biased words by obtaining the
        words closest to each word in the bias_representation set (in case of gender
        "he" and "she").

        Parameters
        ----------
        model: WordEmbeddingModel
            The word embedding model to debias.
        ignore: List[str]
            Set of words to be ignored from the debias process.
        n_words: int
            number of target words to use.
        bias_representation: List[str]
            Two words that represents de bias groups.

        Returns
        -------
        List[str]
            List of target words for each bias group

        """
        similarities = self._bias_by_projection(
            model, target, ignore, bias_representation
        )
        sorted_words = sorted(similarities.items(), key=operator.itemgetter(1))
        female_words = [pair[0] for pair in sorted_words[:n_words]]
        male_words = [pair[0] for pair in sorted_words[-n_words:]]
        return female_words + male_words

    def _principal_components(self, model: WordEmbeddingModel) -> np.ndarray:
        self.pca_type.fit(model.wv.vectors - self.embeddings_mean)
        return self.pca_type.components_

    def _calculate_embeddings_mean(self, model: WordEmbeddingModel) -> float:
        return np.mean(model.wv.vectors, axis=0)

    def _drop_frecuency_features(
        self, components: int, model: WordEmbeddingModel
    ) -> dict[str, np.ndarray]:
        """Remove the frequency features from the embeddings.

        This is done by removing a component from the ones obtain by the PCA over the
        set of embeddings. The component to remove is indicated by parameter
        "components" and it is removed from the target words embeddings.

        Parameters
        ----------
        components: int
            The number of components to be removed.
        model: WordEmbeddingModel
            A word embedding model.

        Returns
        -------
        Dict[str, np.ndarray]
            The new embeddings for the target words.

        """
        droped_frecuencies = {}

        for word in self.target_words:
            embedding = model[word]
            decendecentralize_embedding = embedding - self.embeddings_mean
            frecuency = np.zeros(embedding.shape).astype(float)

            # for u in self.pca[components]:
            u = self.pca[components]
            frecuency = np.dot(np.dot(np.transpose(u), embedding), u)
            new_embedding = decendecentralize_embedding - frecuency

            droped_frecuencies[word] = new_embedding
        return droped_frecuencies

    def _identify_bias_subspace(
        self,
        defining_pairs_embeddings: list[dict[str, np.ndarray]],
        verbose: bool = False,
    ) -> PCA:
        matrix = []
        for embedding_dict_pair in defining_pairs_embeddings:
            # Get the center of the current defining pair.
            pair_embeddings = np.array(list(embedding_dict_pair.values()))
            center = np.mean(pair_embeddings, axis=0)
            # For each word, embedding in the defining pair:
            for embedding in embedding_dict_pair.values():
                # Substract the center of the pair to the embedding
                matrix.append(embedding - center)
        matrix = np.array(matrix)

        pca = PCA(**self.pca_args)
        pca.fit(matrix)

        if verbose:
            explained_variance = pca.explained_variance_ratio_
            print(
                f"PCA variance explained: {explained_variance[0 : pca.n_components_]}"
            )

        return pca

    def _drop(self, u: np.ndarray, v: np.ndarray) -> np.ndarray:
        return u - v * u.dot(v) / v.dot(v)

    def _debias(self, words_dict: dict[str, np.ndarray]) -> dict[str, np.ndarray]:
        for word in tqdm(words_dict):
            embedding = words_dict[word]
            debias_embedding = self._drop(embedding, self.bias_direction)
            words_dict.update({word: debias_embedding})
        return words_dict

    def _get_optimal_dimension(
        self, model: WordEmbeddingModel, n_words: int, n_components: int
    ) -> int:
        n_components = n_components
        scores = []
        for d in range(n_components):
            result_embeddings = self._drop_frecuency_features(d, model)
            result_embeddings = self._debias(result_embeddings)
            y_true = [0] * n_words + [1] * n_words
            scores.append(self._kmeans_eval(result_embeddings, y_true, n_words))
        min_alignment = min(scores)

        return scores.index(min_alignment)

    def _kmeans_eval(
        self,
        embeddings_dict: dict[str, np.ndarray],
        y_true: list[int],
        n_words: int,
        n_cluster: int = 2,
    ) -> float:
        embeddings = [
            embeddings_dict[word] for word in self.target_words[0 : 2 * n_words]
        ]
        kmeans = KMeans(n_cluster).fit(embeddings)
        y_pred = kmeans.predict(embeddings)
        correct = [
            1 if item1 == item2 else 0
            for (item1, item2) in zip(y_true, y_pred, strict=False)
        ]
        alignment_score = sum(correct) / float(len(correct))
        alignment_score = max(alignment_score, 1 - alignment_score)
        return alignment_score

    def fit(
        self,
        model: WordEmbeddingModel,
        definitional_pairs: list[list[str]],
        bias_representation: list[str],
        **fit_params,
    ) -> BaseDebias:
        """Compute the bias direction and get the principal components of the vectors.

        Parameters
        ----------
        model : WordEmbeddingModel
            The word embedding model to debias.
        definitional_pairs : List[List[str]]
            A sequence of string pairs that will be used to define the bias
            direction. For example, for the case of gender debias, this list
            could be [['woman', 'man'], ['girl', 'boy'], ['she', 'he'],
            ['mother', 'father'], ...].
        bias_representation: List[str]
            Two words that represents each bias group. In case of gender
            "he" and "she".

        Returns
        -------
        BaseDebias
            The debias method fitted.

        """
        self.definitional_pairs = definitional_pairs

        self._check_sets_sizes(self.definitional_pairs, "definitional", set_size=2)

        # -------------------------------------------------------------------
        # Obtain the embedding of each definitional pairs.
        if self.verbose:
            print("Obtaining definitional pairs.")

        self.definitional_pairs_embeddings = get_embeddings_from_tuples(
            model=model,
            sets=definitional_pairs,
            sets_name="definitional",
            warn_lost_sets=self.verbose,
            normalize=False,
            verbose=self.verbose,
        )
        # -------------------------------------------------------------------
        # Check bias representation
        if bias_representation[0] not in model or bias_representation[1] not in model:
            raise Exception("bias_representation words not in model")
        self.bias_representation = bias_representation

        # -------------------------------------------------------------------
        # Identify the bias subspace using the defining pairs.
        if self.verbose:
            print("Identifying the bias subspace.")
        self.bias_direction = self._identify_bias_subspace(
            self.definitional_pairs_embeddings, verbose=self.verbose
        ).components_[0]

        # -------------------------------------------------------------------
        # Obtain embeddings' mean
        self.embeddings_mean = self._calculate_embeddings_mean(model)

        # -------------------------------------------------------------------
        # Obtain the principal components of all vector in the model.
        if self.verbose:
            print("Obtaining principal components")
        self.pca = self._principal_components(model)

        return self

    def transform(
        self,
        model: WordEmbeddingModel,
        target: list[str] | None = None,
        ignore: list[str] | None = [],
        copy: bool = True,
    ) -> WordEmbeddingModel:
        """Execute hard debias over the provided model.

        Parameters
        ----------
        model : WordEmbeddingModel
            The word embedding model to debias.
        ignore :  List[str], optional
            If set of words is specified in ignore, the debias
            method will perform the debias in all target words except
            those specified in this set, by default [].
        copy : bool, optional
            If `True`, the debias will be performed on a copy of the
            model.
            If `False`, the debias will be applied on the same model delivered,
            causing its vectors to mutate.
            **WARNING:** Setting copy with `True` requires RAM at least 2x of
            the size of the model, otherwise the execution of the debias may
            raise to `MemoryError`, by default True.

        Returns
        -------
        WordEmbeddingModel
            The debiased embedding model.

        """
        # check if the following attributes exist in the object.
        self._check_transform_args(
            model=model,
            ignore=ignore,
            copy=copy,
        )
        check_is_fitted(
            self,
            [
                "definitional_pairs",
                "definitional_pairs_embeddings",
                "bias_direction",
                "embeddings_mean",
                "pca",
                "bias_representation",
            ],
        )
        if self.verbose:
            print(f"Executing Double Hard Debias on {model.name}")
        # -------------------------------------------------------------------
        # Copy
        if copy:
            print(
                "Copy argument is True. Transform will attempt to create a copy "
                "of the original model. This may fail due to lack of memory."
            )
            model = deepcopy(model)
            print("Model copy created successfully.")

        else:
            print(
                "copy argument is False. The execution of this method will mutate "
                "the original model."
            )
        # -------------------------------------------------------------------
        # Obtain words to apply debias
        if self.verbose:
            print("Obtaining words to apply debias")

        if target:
            self.n_words = len(target) // 2

        self.target_words = self.get_target_words(
            model, target, ignore, self.n_words, self.bias_representation
        )

        # -------------------------------------------------------------------
        # Searching best component of pca to debias
        if self.verbose:
            print("Searching component to debias")
        optimal_dimensions = self._get_optimal_dimension(
            model, self.n_words, self.n_components
        )

        # -------------------------------------------------------------------
        # Execute debias
        if self.verbose:
            print("Executing debias")
        debiased_embeddings = self._drop_frecuency_features(optimal_dimensions, model)
        debiased_embeddings = self._debias(debiased_embeddings)

        # -------------------------------------------------------------------
        # Update vectors
        if self.verbose:
            print("Updating debiased vectors")
        for word in tqdm(debiased_embeddings):
            model.update(word, debiased_embeddings[word].astype(model.wv.vectors.dtype))
        # -------------------------------------------------------------------
        # # Generate the new KeyedVectors
        if self.criterion_name_ is None:
            new_model_name = f"{model.name}_debiased"
        else:
            new_model_name = f"{model.name}_{self.criterion_name_}_debiased"
        model.name = new_model_name
        if self.verbose:
            print("Done!")
        return model



"""Unit tests for the WordEmbeddingModel class from wefe.word_embedding_model."""

import gensim
from gensim.models import FastText, KeyedVectors, Word2Vec
from gensim.test.utils import common_texts
import numpy as np
import pytest

from wefe.word_embedding_model import GENSIM_V4_OR_GREATER, WordEmbeddingModel


@pytest.fixture
def test_keyed_vectors() -> KeyedVectors:
    """Fixture that loads a pre-trained KeyedVectors model for testing."""
    test_model = KeyedVectors.load("./wefe/datasets/data/test_model.kv")
    return test_model


def test__init__() -> None:
    """Test the constructor of the WordEmbeddingModel class.

    It verifies type validations for `wv`, `name`, and `vocab_prefix`
    arguments, and checks for correct initialization of model attributes.
    """
    # target sets None
    with pytest.raises(
        TypeError, match="wv must be an instance of gensim's BaseKeyedVectors, but got"
    ):
        WordEmbeddingModel(wv=None)

    # target sets int
    with pytest.raises(
        TypeError,
        match=(
            "wv must be an instance of gensim's BaseKeyedVectors, but got"
            " <class 'str'>."
        ),
    ):
        WordEmbeddingModel(wv="abc")

    with pytest.raises(
        TypeError,
        match=(
            "wv must be an instance of gensim's BaseKeyedVectors, but got"
            " <class 'int'>."
        ),
    ):
        WordEmbeddingModel(wv=1)

    with pytest.raises(
        TypeError,
        match=(
            "wv must be an instance of gensim's BaseKeyedVectors, but got "
            "<class 'dict'>."
        ),
    ):
        WordEmbeddingModel(wv={})

    w2v = KeyedVectors.load("./wefe/datasets/data/test_model.kv")

    with pytest.raises(
        TypeError,
        match="name must be a string or None, but got <class 'int'>.",
    ):
        WordEmbeddingModel(wv=w2v, name=1)

    with pytest.raises(
        TypeError,
        match=r"vocab_prefix must be a string or None, but got <class 'int'>.",
    ):
        WordEmbeddingModel(wv=w2v, vocab_prefix=1)

    # test models
    model = WordEmbeddingModel(wv=w2v)
    assert model.wv == w2v
    assert model.name == "Unnamed model"
    assert model.vocab_prefix is None

    model = WordEmbeddingModel(wv=w2v, name="w2v_sm")
    assert model.wv == w2v
    assert model.name == "w2v_sm"
    assert model.vocab_prefix is None

    model = WordEmbeddingModel(wv=w2v, name="w2v_sm", vocab_prefix="\\c\\en")
    assert model.wv == w2v
    assert model.name == "w2v_sm"
    assert model.vocab_prefix == "\\c\\en"


def test__eq__(test_keyed_vectors: gensim.models.KeyedVectors) -> None:
    """Test the equality comparison (__eq__) operator for WordEmbeddingModel.

    It checks if two WordEmbeddingModel instances are equal based on their internal
    KeyedVectors object, name, and vocab_prefix.
    """
    model_1 = WordEmbeddingModel(wv=test_keyed_vectors, name="w2v")
    model_2 = WordEmbeddingModel(wv=test_keyed_vectors, name="w2v_2")
    model_3_prefix_a = WordEmbeddingModel(
        wv=test_keyed_vectors, name="w2v_3", vocab_prefix="a"
    )
    model_3_prefix_b = WordEmbeddingModel(
        wv=test_keyed_vectors, name="w2v_3", vocab_prefix="b"
    )

    assert model_1 != ""

    assert model_1 == model_1
    assert model_1 != model_2

    model_1.wv = None

    assert model_1 != model_2

    assert model_2 != model_3_prefix_a
    assert model_3_prefix_a == model_3_prefix_a

    assert model_3_prefix_b != model_3_prefix_a


def test__contains__(test_keyed_vectors: gensim.models.KeyedVectors) -> None:
    """Test the __contains__ operator for WordEmbeddingModel.

    It verifies if a word exists in the model's vocabulary.
    """
    model = WordEmbeddingModel(wv=test_keyed_vectors, name="w2v")

    assert "men" in model
    assert "asdf" not in model
    assert None not in model
    assert 0 not in model


def test__getitem__(test_keyed_vectors: gensim.models.KeyedVectors) -> None:
    """Test the __getitem__ operator for WordEmbeddingModel.

    It checks if a word embedding can be retrieved correctly
    and raises KeyError for words not in the vocabulary.
    """
    model = WordEmbeddingModel(wv=test_keyed_vectors, name="w2v")

    embedding = model["career"]
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (300,)

    with pytest.raises(KeyError, match=r"word 'not_in_vocab' not in model vocab."):
        model["not_in_vocab"]


def test__repr__(test_keyed_vectors: gensim.models.KeyedVectors) -> None:
    """Test the string representation (__repr__) of WordEmbeddingModel.

    It verifies that the representation correctly displays model name,
    vocabulary size, embedding dimensions, and vocab prefix when present.
    """
    model_1 = WordEmbeddingModel(wv=test_keyed_vectors, name="w2v")
    model_1_no_name = WordEmbeddingModel(wv=test_keyed_vectors)
    model_1_prefix_a = WordEmbeddingModel(
        wv=test_keyed_vectors, name="w2v", vocab_prefix="a"
    )
    model_1_no_name_prefix_a = WordEmbeddingModel(
        wv=test_keyed_vectors, vocab_prefix="a"
    )

    assert (
        model_1.__repr__()
        == "<WordEmbeddingModel named 'w2v' with 13013 word embeddings of 300 dims>"
    )
    assert model_1_no_name.__repr__() == (
        "<WordEmbeddingModel 'Unnamed model' with 13013 word embeddings of 300 dims>"
    )
    assert model_1_prefix_a.__repr__() == (
        "<WordEmbeddingModel named 'w2v' with 13013 word embeddings of 300 dims and"
        " 'a' as word prefix>"
    )
    assert model_1_no_name_prefix_a.__repr__() == (
        "<WordEmbeddingModel 'Unnamed model' with 13013 word embeddings of 300 dims "
        "and 'a' as word prefix>"
    )

    del model_1.name
    assert model_1.__repr__() == "<WordEmbeddingModel with wrong __repr__>"


def test__init__with_w2v_model() -> None:
    """Test initialization of WordEmbeddingModel with a gensim.Word2Vec model.

    Ensures that the internal wv attribute correctly points to the Word2Vec's
    KeyedVectors.
    """
    if GENSIM_V4_OR_GREATER:
        w2v = Word2Vec(
            sentences=common_texts, vector_size=100, window=5, min_count=1, workers=-1
        )
    else:
        w2v = Word2Vec(
            sentences=common_texts, size=100, window=5, min_count=1, workers=-1
        )
    w2v_keyed_vectors = w2v.wv
    wem = WordEmbeddingModel(wv=w2v_keyed_vectors, name="w2v")

    assert w2v.wv == wem.wv


def test__init_with_fast_model() -> None:
    """Test initialization of WordEmbeddingModel with a gensim.FastText model.

    Ensures that the internal wv attribute correctly points to the FastText's
    KeyedVectors.
    """
    if GENSIM_V4_OR_GREATER:
        fast = FastText(
            vector_size=4, window=3, min_count=1, sentences=common_texts, epochs=10
        )
    else:
        fast = FastText(size=4, window=3, min_count=1, sentences=common_texts, iter=10)
    fast_keyed_vectors = fast.wv
    wem = WordEmbeddingModel(wv=fast_keyed_vectors, name="w2v")

    assert fast.wv == wem.wv


def test_normalize_embeddings(test_keyed_vectors: gensim.models.KeyedVectors) -> None:
    """Test the normalize method of WordEmbeddingModel.

    It checks if embeddings are correctly normalized to unit length and
    handles cases where the internal gensim model cannot be normalized.
    """
    model = WordEmbeddingModel(wv=test_keyed_vectors, name="w2v")

    # test unnormalized embeddings
    for word in model.vocab:
        assert np.abs(np.linalg.norm(model[word]) - 1.0) > 0.000001

    # test normalized embeddings
    model.normalize()
    for word in model.vocab:
        assert np.linalg.norm(model[word]) - 1.0 < 0.000001

    model.wv = None

    with pytest.raises(
        AttributeError,
        match=(
            r"The underlying gensim model does not have a known normalization method "
            r"\('get_normed_vectors' or 'init_sims'\)."
        ),
    ):
        model.normalize()


def test_update_embeddings(
    test_keyed_vectors: gensim.models.KeyedVectors,
) -> None:
    """Test the batch_update method of WordEmbeddingModel.

    It checks if multiple word embeddings can be updated simultaneously
    and validates input types and sizes.
    """
    model = WordEmbeddingModel(wv=test_keyed_vectors, name="w2v")

    words = ["The", "in"]
    embeddings = [
        np.ones(300, dtype=np.float32),
        np.ones(300, dtype=np.float32) * -1,
    ]

    model.batch_update(words=words, embeddings=embeddings)

    assert all(model["The"] == embeddings[0])
    assert all(model["in"] == embeddings[1])

    embeddings_in_array = np.array(
        [
            np.ones(300, dtype=np.float32) * 2,
            np.ones(300, dtype=np.float32) * -2,
        ]
    )
    model.batch_update(words=words, embeddings=embeddings_in_array)

    assert all(model["The"] == embeddings_in_array[0])
    assert all(model["in"] == embeddings_in_array[1])

    # words argument should be a list, tuple or np.array of strings
    with pytest.raises(
        TypeError,
        match=(
            r"words argument should be a list, tuple or np.array of strings, "
            r"but got <class 'NoneType'>."
        ),
    ):
        model.batch_update(words=None, embeddings=embeddings)

    # embeddings argument should be a list, tuple or np.array of NumPy arrays
    with pytest.raises(
        TypeError,
        match=(
            r"embeddings argument should be a list, tuple or np.array of NumPy arrays, "
            r"but got <class 'NoneType'>."
        ),
    ):
        model.batch_update(words=words, embeddings=None)

    # words and embeddings must have the same size
    with pytest.raises(
        ValueError,
        match=(
            r"words and embeddings must have the same number of elements, "
            "but got 3 words and 2 embeddings."
        ),
    ):
        model.batch_update(words=words + ["is"], embeddings=embeddings)

    # Check for non-string word in words
    with pytest.raises(
        TypeError,
        match=r"All elements in 'words' must be strings, but found a <class 'int'>.",
    ):
        model.batch_update(words=["The", 1], embeddings=embeddings)

    # Check for word not in vocab
    with pytest.raises(
        ValueError,
        match=r"The following words are not in the model's vocabulary: not_in_vocab.",
    ):
        model.batch_update(words=["The", "not_in_vocab"], embeddings=embeddings)

    # Check for non-numpy embedding in embeddings
    with pytest.raises(
        TypeError,
        match=(
            r"Embedding at index 1 \('in'\) is not a NumPy array, but "
            r"got <class 'int'>."
        ),
    ):
        model.batch_update(words=words, embeddings=[embeddings[0], 1])

    # Check for wrong embedding shape
    with pytest.raises(
        ValueError,
        match=(
            r"Embedding at index 1 \('in'\) has shape \(200,\) which is different from "
            r"the model's embedding size \(300,\)."
        ),
    ):
        model.batch_update(
            words=words,
            embeddings=[embeddings[0], np.ones(200, dtype=np.float32)],
        )

    # Check for wrong embedding dtype
    with pytest.raises(
        ValueError,
        match=(
            r"Embedding at index 1 \('in'\) with dtype \(float64\) cannot be safely "
            r"cast to model's dtype \(float32\)."
        ),
    ):
        model.batch_update(
            words=words,
            embeddings=[embeddings[0], np.ones(300, dtype=np.float64)],
        )



"""Contains a base class for implement any debias method in WEFE."""

from abc import abstractmethod

from sklearn.base import BaseEstimator

from wefe.word_embedding_model import WordEmbeddingModel


class BaseDebias(BaseEstimator):
    """Mixin class for implement any debias method in WEFE."""

    # The name of the method.
    name: str

    @abstractmethod
    def fit(
        self,
        model: WordEmbeddingModel,
        **fit_params,
    ) -> "BaseDebias":
        """Fit the transformation.

        Parameters
        ----------
        model : WordEmbeddingModel
            The word embedding model to debias.
        verbose: bool, optional
            True will print informative messages about the debiasing process,
            by default False.

        """
        raise NotImplementedError()

    @abstractmethod
    def transform(
        self,
        model: WordEmbeddingModel,
        target: list[str] | None = None,
        ignore: list[str] | None = None,
        copy: bool = True,
    ) -> WordEmbeddingModel:
        """Perform the debiasing method over the model provided.

        Parameters
        ----------
        model : WordEmbeddingModel
            The word embedding model to debias.
        target : Optional[List[str]], optional
            If a set of words is specified in target, the debias method will be
            performed only on the word embeddings of this set. In the case of provide
            `None`, the debias will be performed on all words (except those specified
            in ignore).
            by default `None`.
        ignore : Optional[List[str]], optional
            If target is `None` and a set of words is specified in ignore, the debias
            method will perform the debias in all words except those specified in this
            set, by default `None`.
        copy : bool, optional
            If `True`, the debias will be performed on a copy of the model.
            If `False`, the debias will be applied on the same model delivered, causing
            its vectors to mutate.
            **WARNING:** Setting copy with `True` requires at least 2x RAM of the size
            of the model. Otherwise the execution of the debias may raise
            `MemoryError`, by default True.

        Returns
        -------
        WordEmbeddingModel
            The debiased word embedding model.

        """
        raise NotImplementedError()

    def fit_transform(
        self,
        model: WordEmbeddingModel,
        target: list[str] | None = None,
        ignore: list[str] | None = None,
        copy: bool = True,
        **fit_params,
    ) -> WordEmbeddingModel:
        """Convenience method to execute fit and transform in a single call.

        Parameters
        ----------
        model : WordEmbeddingModel
            A word embedding model object.
        target : Optional[List[str]], optional
            If a set of words is specified in target, the debias method will be applied
            only on the word embeddings of this set, by default None.
        ignore : Optional[List[str]], optional
            If target is None and a set of words is specified in ignore, the debias
            method will debias all words except those specified in ignore,
            by default None.
        copy : bool, optional
            If True, the debias will be performed on a copy of the model.
            If False, the debias will be applied on the same model delivered, causing
            its vectors to mutate.
            **WARNING:** Setting copy with True requires at least 2x RAM of the size
            of the model. Otherwise the execution of the debias may raise
            `MemoryError`, by default True.
        verbose : bool, optional
            True will print informative messages about the debiasing process,
            by default True.

        Returns
        -------
        WordEmbeddingModel
            The debiased word embedding model.

        """
        return self.fit(model, **fit_params).transform(
            model, target=target, ignore=ignore, copy=copy
        )

    def _check_transform_args(
        self,
        model: WordEmbeddingModel,
        target: list[str] | None = None,
        ignore: list[str] | None = None,
        copy: bool = True,
    ) -> None:
        # check if model is a WordEmbeddingModel
        if not isinstance(model, WordEmbeddingModel):
            raise TypeError(
                f"model should be a WordEmbeddingModel instance, got {model}."
            )

        # check if target is a list
        if target is not None and not isinstance(target, list):
            raise TypeError(
                f"target should be None or a list of strings, got {target}."
            )

        # check if all elements of the list are strings
        if isinstance(target, list):
            for idx, word in enumerate(target):
                if not isinstance(word, str):
                    raise TypeError(
                        "All elements in target should be strings"
                        f", got: {word} at index {idx} "
                    )

        # check ignore parameter type (none or list)
        if ignore is not None and not isinstance(ignore, list):
            raise TypeError(
                f"ignore should be None or a list of strings, got {ignore}."
            )

        # if ignore is a list, check that all its elements are string
        if isinstance(ignore, list):
            for idx, word in enumerate(ignore):
                if not isinstance(word, str):
                    raise TypeError(
                        "All elements in ignore should be strings"
                        f", got: {word} at index {idx} "
                    )

        # check copy
        if not isinstance(copy, bool):
            raise TypeError(f"copy should be a bool, got {copy}.")

    def _check_sets_sizes(
        self,
        sets: list[list[str]],
        set_name: str,
        set_size: int | str,
    ) -> None:
        if len(sets) == 0:
            raise ValueError("")

        # case fixed set_size.
        if isinstance(set_size, int):
            for idx, set_ in enumerate(sets):
                if len(set_) != set_size:
                    adverb = "less" if len(set_) < set_size else "more"

                    raise ValueError(
                        f"The {set_name} set at position {idx} ({set_}) has {adverb} "
                        f"words than allowed by {self.name}: "
                        f"got {len(set_)} words, expected {set_size}."
                    )

        # case free set_size.
        elif set_size == "n":
            inferred_set_size = len(sets[0])

            for idx, set_ in enumerate(sets):
                if len(set_) != inferred_set_size:
                    adverb = "less" if len(set_) < inferred_set_size else "more"

                    raise ValueError(
                        f"The {set_name} set at position {idx} ({set_}) has {adverb} "
                        f"words than the other {set_name} sets: "
                        f"got {len(set_)} words, expected {inferred_set_size}."
                    )

        else:
            raise ValueError('Wrong set_size value {set_size}. Expected int or "n"')



"""Multiclass Hard Debias WEFE implementation."""

from copy import deepcopy
import logging
from typing import Any

import numpy as np
from sklearn.decomposition import PCA
from tqdm import tqdm

from wefe.debias.base_debias import BaseDebias
from wefe.preprocessing import get_embeddings_from_tuples
from wefe.utils import check_is_fitted
from wefe.word_embedding_model import EmbeddingDict, WordEmbeddingModel

logger = logging.getLogger(__name__)


class MulticlassHardDebias(BaseDebias):
    """Generalized version of Hard Debias that enables multiclass debiasing.

    Generalized refers to the fact that this method extends Hard Debias in order to
    support more than two types of social target sets within the definitional set.
    For example, for the case of religion bias, it supports a debias using words
    associated with Christianity, Islam and Judaism.

    Examples
    --------
    .. note::

        For more information on the use of mitigation methods, visit
        :ref:`bias mitigation` in the User Guide.

    The following example shows how to run an ethnicity debias based on Black,
    Asian and Caucasian groups.

    >>> from wefe.datasets import fetch_debias_multiclass, load_weat
    >>> from wefe.debias.multiclass_hard_debias import MulticlassHardDebias
    >>> from wefe.utils import load_test_model
    >>>
    >>> model = load_test_model()  # load a reduced version of word2vec
    >>>
    >>> # obtain the sets of words that will be used in the debias process.
    >>> multiclass_debias_wordsets = fetch_debias_multiclass()
    >>> weat_wordsets = load_weat()
    >>>
    >>> ethnicity_definitional_sets = (
    ...     multiclass_debias_wordsets["ethnicity_definitional_sets"]
    ... )
    >>> ethnicity_equalize_sets = list(
    ...     multiclass_debias_wordsets["ethnicity_analogy_templates"].values()
    ... )
    >>>
    >>> # instance the debias object that will perform the mitigation
    >>> mhd = MulticlassHardDebias(verbose=False, criterion_name="ethnicity")
    >>> # fits the transformation parameters (bias direction, etc...)
    >>> mhd.fit(
    ...     model=model,
    ...     definitional_sets=ethnicity_definitional_sets,
    ...     equalize_sets=ethnicity_equalize_sets,
    ... )
    >>>
    >>> # perform the transformation (debiasing) on the embedding model
    >>> ethnicity_debiased_model = mhd.transform(model, copy=True)

    References
    ----------
    | [1]: Manzini, T., Chong, L. Y., Black, A. W., & Tsvetkov, Y. (2019, June).
    | Black is to Criminal as Caucasian is to Police: Detecting and Removing Multiclass
    | Bias in Word Embeddings.
    | In Proceedings of the 2019 Conference of the North American Chapter of the
    | Association for Computational Linguistics: Human Language Technologies,
    | Volume 1 (Long and Short Papers) (pp. 615-621).
    | [2]: https://github.com/TManzini/DebiasMulticlassWordEmbedding

    """

    def __init__(
        self,
        pca_args: dict[str, Any] = {"n_components": 10},
        verbose: bool = False,
        criterion_name: str | None = None,
    ) -> None:
        """Initialize a Multiclass Hard Debias instance.

        Parameters
        ----------
        pca_args : Dict[str, Any], optional
            Arguments for the PCA that is calculated internally in the identification
            of the bias subspace, by default {"n_components": 10}
        verbose : bool, optional
            True will print informative messages about the debiasing process,
            by default False.
        criterion_name : Optional[str], optional
            The name of the criterion for which the debias is being executed,
            e.g. 'Gender'. This will indicate the name of the model returning transform,
            by default None

        """
        # check pca args
        if not isinstance(pca_args, dict):
            raise TypeError(f"pca_args should be a dict, got {pca_args}.")
        # check verbose
        if not isinstance(verbose, bool):
            raise TypeError(f"verbose should be a bool, got {verbose}.")

        self.pca_args = pca_args
        self.verbose = verbose

        self.pca_num_components_ = pca_args.get("n_components", 10)

        if criterion_name is None or isinstance(criterion_name, str):
            self.criterion_name_ = criterion_name
        else:
            raise ValueError(
                f"debias_criterion_name should be str, got: {criterion_name}"
            )

    def _identify_bias_subspace(
        self,
        definning_sets_embeddings: list[EmbeddingDict],
    ) -> PCA:
        matrix = []
        for definning_set_dict in definning_sets_embeddings:
            # Get the center of the current definning pair.
            set_embeddings = np.array(list(definning_set_dict.values()))
            center = np.mean(set_embeddings, axis=0)
            # For each word, embedding in the definning pair:
            for embedding in definning_set_dict.values():
                # Substract the center of the pair to the embedding
                matrix.append(embedding - center)
        matrix = np.array(matrix)  # type: ignore

        pca = PCA(**self.pca_args)
        pca.fit(matrix)

        if self.verbose:
            explained_variance = pca.explained_variance_ratio_
            if len(explained_variance) > 10:
                logger.info(f"PCA variance explaned: {explained_variance[0:10]}")
            else:
                logger.info(f"PCA variance explaned: {explained_variance}")

        return pca

    def _project_onto_subspace(
        self,
        vector: np.ndarray,
        subspace: np.ndarray,
    ) -> np.ndarray:
        v_b = np.zeros_like(vector)
        for component in subspace:
            v_b += np.dot(vector.transpose(), component) * component
        return v_b

    def _neutralize(
        self,
        model: WordEmbeddingModel,
        bias_subspace: np.ndarray,
        target: list[str] | None,
        ignore: list[str] | None,
    ) -> None:
        target_ = set(target) if target is not None else set(model.vocab.keys())

        ignore_ = set(ignore) if ignore is not None and target is None else set()

        # filtering the defitional sets in order to ignore these words in the
        # neutralization.
        definitional_words = set(np.array(self.definitional_sets_).flatten().tolist())
        target_ = target_.difference(definitional_words)

        for word in tqdm(target_):
            if word not in ignore_:
                # get the embedding
                v = model[word]
                # neutralize the embedding if the word is not in the definitional words.
                v_b = self._project_onto_subspace(v, bias_subspace)
                # neutralize the embedding
                new_v = (v - v_b) / np.linalg.norm(v - v_b)
                # update the old values
                model.update(word, new_v)

    def _equalize(
        self,
        model: WordEmbeddingModel,
        equalize_sets_embeddings: list[EmbeddingDict],
        bias_subspace: np.ndarray,
    ) -> None:
        for equalize_tuple_embeddings in equalize_sets_embeddings:
            words = equalize_tuple_embeddings.keys()
            embeddings = np.array(list(equalize_tuple_embeddings.values()))

            # calculate the mean of the equality set
            mean = np.mean(embeddings, axis=0)
            # project the mean in the bias subspace
            mean_b = self._project_onto_subspace(mean, bias_subspace)
            # discard the projection from the mean
            upsilon = mean - mean_b

            for word, embedding in zip(words, embeddings, strict=False):
                v_b = self._project_onto_subspace(embedding, bias_subspace)
                frac = (v_b - mean_b) / np.linalg.norm(v_b - mean_b)
                new_v = upsilon + np.sqrt(1 - np.sum(np.square(upsilon))) * frac
                model.update(word, new_v)

    def fit(
        self,
        model: WordEmbeddingModel,
        definitional_sets: list[list[str]],
        equalize_sets: list[list[str]],
    ) -> BaseDebias:
        """Compute the bias direction and obtains the equalize embedding pairs.

        Parameters
        ----------
        model : WordEmbeddingModel
            The word embedding model to debias.
        definitional_sets : List[List[str]]
            A sequence of string pairs that will be used to define the bias direction.
            For example, for the case of gender debias, this list could be [['woman',
            'man'], ['girl', 'boy'], ['she', 'he'], ['mother', 'father'], ...].
            Multiclass hard debias also accepts lists of sets of more than two words,
            such as religion where sets of words representing Christianity, Islam and
            Judaism can be used. See the example for more information.
        equalize_pairs : Optional[List[List[str]]], optional
            A list with pairs of strings, which will be equalized.
            In the case of passing None, the equalization will be done over the word
            pairs passed in definitional_sets,
            by default None.

        Returns
        -------
        BaseDebias
            The debias method fitted.

        """
        # -----------------------------------------------------------------------------:
        # Obtain the embedding of the definitional sets.
        self._check_sets_sizes(definitional_sets, set_name="definitional", set_size="n")

        if self.verbose:
            print("Obtaining definitional sets.")
        self.definitional_sets_ = definitional_sets
        self.definitional_sets_embeddings_ = get_embeddings_from_tuples(
            model=model,
            sets=definitional_sets,
            sets_name="definitional",
            warn_lost_sets=True,
            normalize=True,
            verbose=self.verbose,
        )

        # -----------------------------------------------------------------------------:
        # Identify the bias subspace using the definning sets.
        if self.verbose:
            print("Identifying the bias subspace.")
        self.pca_ = self._identify_bias_subspace(
            self.definitional_sets_embeddings_,
        )
        self.bias_subspace_ = self.pca_.components_[: self.pca_num_components_]

        # ------------------------------------------------------------------------------
        # Equalize embeddings:

        # Get the equalization sets embeddings.
        # Note that the equalization sets are the same as the definitional sets.
        if self.verbose:
            print("Obtaining equalize pairs.")
        self.equalize_sets_embeddings_ = get_embeddings_from_tuples(
            model=model,
            sets=equalize_sets,
            sets_name="equalize",
            normalize=True,
            warn_lost_sets=True,
            verbose=self.verbose,
        )
        return self

    def transform(
        self,
        model: WordEmbeddingModel,
        target: list[str] | None = None,
        ignore: list[str] | None = None,
        copy: bool = True,
    ) -> WordEmbeddingModel:
        """Execute Multiclass Hard Debias over the provided model.

        Parameters
        ----------
        model : WordEmbeddingModel
            The word embedding model to debias.
        target : Optional[List[str]], optional
            If a set of words is specified in target, the debias method will be
            performed only on the word embeddings of this set. If `None` is provided,
            the debias will be performed on all words (except those specified in
            ignore).
            Note that some words that are not in target may be modified due to the
            equalization process.
            by default `None`.
        ignore : Optional[List[str]], optional
            If target is `None` and a set of words is specified in ignore, the debias
            method will perform the debias in all words except those specified in this
            set.
            Note that some words that are in ignore may be modified due to the
            equalization process.
            by default `None`.
        copy : bool, optional
            If `True`, the debias will be performed on a copy of the model.
            If `False`, the debias will be applied on the same model delivered, causing
            its vectors to mutate.
            **WARNING:** Setting copy with `True` requires RAM at least 2x of the size
            of the model, otherwise the execution of the debias may raise to
            `MemoryError`, by default True.

        Returns
        -------
        WordEmbeddingModel
            The debiased embedding model.

        """
        self._check_transform_args(
            model=model,
            target=target,
            ignore=ignore,
            copy=copy,
        )

        check_is_fitted(
            self,
            [
                "definitional_sets_",
                "definitional_sets_embeddings_",
                "pca_",
                "bias_subspace_",
            ],
        )

        if self.verbose:
            print(f"Executing Multiclass Hard Debias on {model.name}")

        # ------------------------------------------------------------------------------
        # Copy
        if copy:
            print(
                "copy argument is True. Transform will attempt to create a copy "
                "of the original model. This may fail due to lack of memory."
            )
            model = deepcopy(model)
            print("Model copy created successfully.")

        else:
            print(
                "copy argument is False. The execution of this method will mutate "
                "the original model."
            )

        # ------------------------------------------------------------------------------
        # Neutralize the embeddings provided in target.
        # if target is None, the debias will be performed in all embeddings.
        if self.verbose:
            print("Normalizing embeddings.")
        model.normalize()

        # Neutralize the embeddings:
        if self.verbose:
            print("Neutralizing embeddings")

        # get the words that will be debiased.
        self._neutralize(
            model=model,
            bias_subspace=self.bias_subspace_,
            target=target,
            ignore=ignore,
        )

        if self.verbose:
            print("Normalizing embeddings.")
        model.normalize()

        # ------------------------------------------------------------------------------
        # Equalize embeddings:

        # Execute the equalization
        logger.debug("Equalizing embeddings..")
        self._equalize(
            model=model,
            equalize_sets_embeddings=self.equalize_sets_embeddings_,
            bias_subspace=self.bias_subspace_,
        )

        # ------------------------------------------------------------------------------
        if self.verbose:
            print("Normalizing embeddings.")
        model.normalize()

        # ------------------------------------------------------------------------------
        # # Generate the new KeyedVectors
        if self.criterion_name_ is None:
            new_model_name = f"{model.name}_debiased"
        else:
            new_model_name = f"{model.name}_{self.criterion_name_}_debiased"
        model.name = new_model_name

        if self.verbose:
            print("Done!")

        return model



"""Word Embedding Assosiation Test (WEAT) metric implementation."""

from collections.abc import Callable
import logging
import math
from typing import Any

import numpy as np
from sklearn.metrics.pairwise import cosine_similarity

from wefe.metrics.base_metric import BaseMetric
from wefe.preprocessing import get_embeddings_from_query
from wefe.query import Query
from wefe.word_embedding_model import EmbeddingDict, WordEmbeddingModel


def _left_sided_test_function(calculated: np.number, original: np.number) -> bool:
    return calculated < original


def _right_sided_test_function(calculated: np.number, original: np.number) -> bool:
    return calculated > original


def _two_sided_test_function(calculated: np.number, original: np.number) -> bool:
    return np.abs(calculated) > original


TEST_FUNCTION_DISPATCHER = {
    "left-sided": _left_sided_test_function,
    "right-sided": _right_sided_test_function,
    "two-sided": _two_sided_test_function,
}


class WEAT(BaseMetric):
    r"""Word Embedding Association Test (WEAT).

    The following description of the metric is WEFE's adaptation of what was presented
    in the original WEAT work "Semantics derived automatically from language corpora
    contain human-like biases" [1].

    WEAT receives two sets :math:`T_1` and :math:`T_2` of target words,
    and two sets :math:`A_1` and :math:`A_2` of attribute words and performs a
    hypothesis test on the following null hypothesis:
    There is no difference between the two sets of target words in terms of their
    relative similarity to the similarity with the two sets of attribute words.

    In formal terms, let :math:`T_1` and :math:`T_2` be two sets of target words of
    equal size, and :math:`A_1`, :math:`A_2` the two sets of attribute words.
    Let :math:`\cos(\vec{a}, \vec{b})` denote the cosine of the angle between the
    vectors :math:`\vec{a}` and :math:`\vec{b}`. The test statistic is:

    .. math::

        \text{WEAT}(T_1,T_2,A_1,A_2) = \sum_{x \in T_1} s(x, A_1, A_2) -
        \sum_{y \in T_2} s(y, A_1, A_2)

    where

    .. math::

        s(w, A, B)=\text{mean}_{a \in A} \cos(\vec{w}, \vec{a}) -
        \text{mean}_{b \in B} \cos(\vec{w},\vec{b})

    :math:`s(w,A,B)` measures the association of :math:`w` with the
    attributes, and :math:`\text{WEAT}(T_1,T_2,A_1,A_2)` measures the differential
    association of the two sets of target words with the attribute.

    This metric also contains a variant: WEAT Effect Size (WEAT-ES). This variant
    represents a normalized measure that quantifies how far apart the two distributions
    of association between targets and attributes are. In practical terms, WEAT
    Effect Size makes the metric not dependent on the number of words used in each set.

    .. math::

        \text{WEAT-ES}(T_1,T_2,A_1,A_2) = \frac{\text{mean}_{x \in T_1}\,
        s(x, A_1, A_2) - \text{mean}_{y \in T_2}\, s(y, A_1, A_2) }
        {\text{std-dev}_{w \in T_1 \cup T_2}\, s(w, A_1, A_2)}

    The permutation test measures the (un)likelihood of the null hypothesis by
    computing the probability that a random permutation of the attribute words would
    produce the observed (or greater) difference in sample mean.

    Let :math:`{(T_{1_i},T_{2_i})}_{i}` denote all the partitions of
    :math:`T_1 \cup T_2` into two sets of equal size. The one-sided p-value of the
    permutation test is:

    .. math::

        \text{Pr}_{i}[s(T_{1_i}, T_{2_i}, A_1, A_2) > s(T_1, T_2, A_1, A_2)]

    References
    ----------
    | [1]: Aylin Caliskan, Joanna J Bryson, and Arvind Narayanan.
    |      Semantics derived automatically from language corpora contain human-like
           biases.
    |      Science, 356(6334):183–186, 2017.

    """

    metric_template = (2, 2)
    metric_name = "Word Embedding Association Test"
    metric_short_name = "WEAT"

    def _calc_s(
        self,
        w: np.ndarray,
        A: np.ndarray,
        B: np.ndarray,
    ) -> np.number:
        A_mean_sim = np.mean(cosine_similarity([w], A), dtype=np.float64)
        B_mean_sim = np.mean(cosine_similarity([w], B), dtype=np.float64)
        return A_mean_sim - B_mean_sim

    def _calc_weat(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        A: np.ndarray,
        B: np.ndarray,
    ) -> np.number:
        first_term = np.sum([self._calc_s(x, A, B) for x in X], dtype=np.float64)
        second_term = np.sum([self._calc_s(y, A, B) for y in Y], dtype=np.float64)
        return first_term - second_term

    def _calc_effect_size(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        A: np.ndarray,
        B: np.ndarray,
    ) -> np.number:
        first_term = np.mean([self._calc_s(x, A, B) for x in X], dtype=np.float64)
        second_term = np.mean([self._calc_s(y, A, B) for y in Y], dtype=np.float64)

        std_dev = np.std(
            [self._calc_s(w, A, B) for w in np.concatenate((X, Y))], dtype=np.float64
        )

        return (first_term - second_term) / std_dev

    def _calc_p_value(
        self,
        target_embeddings: list[EmbeddingDict],
        attribute_embeddings: list[EmbeddingDict],
        original_score: np.number,
        iterations: int,
        method: str,
        test_type: str,
        verbose: bool,
    ) -> np.number:
        # TODO: Add joblib or other library to parallelize this function.
        # TODO: Test this function
        # TODO: Implement exact and bootstrap methods
        # TODO: Refactor this function to be able to work with other metrics.
        # maybe this function could be extended from the basemetric class.

        if verbose:
            logging.info(f"weat_original_result {original_score}")

        if method == "exact":
            raise NotImplementedError
        if method != "approximate":
            raise Exception(
                f'p value method should be "exact", "approximate", got {method}.'
            )

        if test_type not in ["left-sided", "right-sided", "two-sided"]:
            raise Exception(
                f'p value test_type should be "left-sided", "right-sided" '
                f'or "two-sided", got {test_type}'
            )
        test_function = TEST_FUNCTION_DISPATCHER[test_type]

        if not isinstance(iterations, int | float):
            raise TypeError(
                f"p value iterations should be int instance, got {iterations}."
            )

        if not isinstance(verbose, bool):
            raise TypeError(f"verbose should be bool instance, got {verbose}.")

        # get attribute embeddings
        attribute_0 = np.array(list(attribute_embeddings[0].values()))
        attribute_1 = np.array(list(attribute_embeddings[1].values()))

        # generate the pool of target and attribute embeddings.
        target_0_dict, target_1_dict = target_embeddings
        pool_target_sets = {**target_0_dict, **target_1_dict}
        pool_target_words: list[str] = list(pool_target_sets.keys())

        len_target_0 = len(target_0_dict)
        number_of_target_words = len(pool_target_words)

        total_permutations = math.factorial(number_of_target_words)
        runs = np.min((iterations, total_permutations))

        permutations_seen: set[tuple] = set()

        if verbose:
            logging.info(
                f"Number of possible total permutations: {total_permutations}. "
                f"Maximum iterations allowed: {runs}"
            )

        count_pass_function = 0

        rng = np.random.default_rng()

        while len(permutations_seen) < runs:
            if verbose and len(permutations_seen) % 500 == 0:
                logging.info(f"WEAT p-value: {len(permutations_seen)} / {runs} runs")

            permutation = tuple(
                rng.choice(pool_target_words, size=number_of_target_words).tolist()
            )
            if permutation not in permutations_seen:
                X_i_words = list(permutation[0:len_target_0])
                Y_i_words = list(permutation[len_target_0:])

                X_i = np.array([pool_target_sets[word] for word in X_i_words])
                Y_i = np.array([pool_target_sets[word] for word in Y_i_words])

                curr_permutation_weat = self._calc_weat(
                    X_i, Y_i, attribute_0, attribute_1
                )

                if test_function(curr_permutation_weat, original_score):
                    count_pass_function += 1

                permutations_seen.add(permutation)

        p_value = (count_pass_function + 1) / (
            runs + 1
        )  # to make p-value unbiased estimator

        if verbose:
            logging.info(
                f"Number of runs: {runs}, Permutations that pass the test function "
                f"type: {count_pass_function}, p-value: {p_value}"
            )
        return p_value

    def run_query(
        self,
        query: Query,
        model: WordEmbeddingModel,
        return_effect_size: bool = False,
        calculate_p_value: bool = False,
        p_value_test_type: str = "right-sided",
        p_value_method: str = "approximate",
        p_value_iterations: int = 10000,
        p_value_verbose: bool = False,
        lost_vocabulary_threshold: float = 0.2,
        preprocessors: list[dict[str, str | bool | Callable]] = [{}],
        strategy: str = "first",
        normalize: bool = False,
        warn_not_found_words: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Calculate the WEAT metric over the provided parameters.

        Parameters
        ----------
        query : Query
            A Query object that contains the target and attribute sets to be tested.

        model : WordEmbeddingModel
            A word embedding model.

        return_effect_size : bool, optional
            Specifies if the returned score in 'result' field of results dict
            is by default WEAT effect size metric, by default False

        calculate_p_value : bool, optional
            Specifies whether the p-value will be calculated through a permutation test.
            Warning: This can increase the computing time quite a lot, by default False.

        p_value_test_type : {'left-sided', 'right-sided', 'two-sided}, optional
            When calculating the p-value, specify the type of test to be performed.
            The options are 'left-sided', 'right-sided' and 'two-sided'
            , by default 'right-sided'

        p_value_method : {'exact', 'approximate'}, optional
            When calculating the p-value, specify the method for calculating the
            p-value. This can be 'exact' and 'approximate'.
            by default 'approximate'.

        p_value_iterations : int, optional
            If the p-value is calculated and the chosen method is 'approximate',
            it specifies the number of iterations that will be performed
            , by default 10000.

        p_value_verbose : bool, optional
            In case of calculating the p-value, specify if notification messages
            will be logged during its calculation, by default False.

        lost_vocabulary_threshold : float, optional
            Specifies the proportional limit of words that any set of the query is
            allowed to lose when transforming its words into embeddings.
            In the case that any set of the query loses proportionally more words
            than this limit, the result values will be np.nan, by default 0.2

        preprocessors : List[Dict[str, Union[str, bool, Callable]]]
            A list with preprocessor options.

            A ``preprocessor`` is a dictionary that specifies what processing(s) are
            performed on each word before it is looked up in the model vocabulary.
            For example, the ``preprocessor``
            ``{'lowecase': True, 'strip_accents': True}`` allows you to lowercase
            and remove the accent from each word before searching for them in the
            model vocabulary. Note that an empty dictionary ``{}`` indicates that no
            preprocessing is done.

            The possible options for a preprocessor are:

            *   ``lowercase``: ``bool``. Indicates that the words are transformed to
                lowercase.
            *   ``uppercase``: ``bool``. Indicates that the words are transformed to
                uppercase.
            *   ``titlecase``: ``bool``. Indicates that the words are transformed to
                titlecase.
            *   ``strip_accents``: ``bool``, ``{'ascii', 'unicode'}``: Specifies that
                the accents of the words are eliminated. The stripping type can be
                specified. True uses 'unicode' by default.
            *   ``preprocessor``: ``Callable``. It receives a function that operates
                on each word. In the case of specifying a function, it overrides the
                default preprocessor (i.e., the previous options stop working).

            A list of preprocessor options allows you to search for several
            variants of the words into the model. For example, the preprocessors
            ``[{}, {"lowercase": True, "strip_accents": True}]``
            ``{}`` allows searching first for the original words in the vocabulary of
            the model. In case some of them are not found,
            ``{"lowercase": True, "strip_accents": True}`` is executed on these words
            and then they are searched in the model vocabulary.

        strategy : str, optional
            The strategy indicates how it will use the preprocessed words: 'first' will
            include only the first transformed word found. 'all' will include all
            transformed words found, by default "first".

        normalize : bool, optional
            True indicates that embeddings will be normalized, by default False

        warn_not_found_words : bool, optional
            Specifies if the function will warn (in the logger)
            the words that were not found in the model's vocabulary, by default False.

        Returns
        -------
        Dict[str, Any]
            A dictionary with the query name, the resulting score of the metric,
            and the scores of WEAT and the effect size of the metric.

        Examples
        --------
        The following example shows how to run a query that measures gender
        bias using WEAT:

        >>> from wefe.query import Query
        >>> from wefe.utils import load_test_model
        >>> from wefe.metrics import WEAT
        >>>
        >>> # define the query
        >>> query = Query(
        ...     target_sets=[
        ...         ["female", "woman", "girl", "sister", "she", "her", "hers",
        ...          "daughter"],
        ...         ["male", "man", "boy", "brother", "he", "him", "his", "son"],
        ...     ],
        ...     attribute_sets=[
        ...         ["home", "parents", "children", "family", "cousins", "marriage",
        ...          "wedding", "relatives",
        ...         ],
        ...         ["executive", "management", "professional", "corporation", "salary",
        ...          "office", "business", "career",
        ...         ],
        ...     ],
        ...     target_sets_names=["Female terms", "Male Terms"],
        ...     attribute_sets_names=["Family", "Career"],
        ... )
        >>>
        >>> # load the model (in this case, the test model included in wefe)
        >>> model = load_test_model()
        >>>
        >>> # instance the metric and run the query
        >>> WEAT().run_query(query, model)
        {'query_name': 'Female terms and Male Terms wrt Family and Career',
        'result': 0.4634388245467562,
        'weat': 0.4634388245467562,
        'effect_size': 0.45076532408312986,
        'p_value': nan}

        If you want to return the effect size as result value, use
        `return_effect_size` parameter as `True` while running the query.

        >>> WEAT().run_query(query, model, return_effect_size=True)
        {'query_name': 'Female terms and Male Terms wrt Family and Career',
        'result': 0.45076532408312986,
        'weat': 0.4634388245467562,
        'effect_size': 0.45076532408312986,
        'p_value': nan}

        If you want the embeddings to be normalized before calculating the metrics
        use the `normalize` parameter as `True` before executing the query.

        >>> WEAT().run_query(query, model, normalize=True)
        {'query_name': 'Female terms and Male Terms wrt Family and Career',
        'result': 0.4634388248814503,
        'weat': 0.4634388248814503,
        'effect_size': 0.4507653062895615,
        'p_value': nan}

        Using the `calculate_p_value` parameter as `True` you can indicate WEAT to run
        the permutation test and return its p-value. The argument
        `p_value_method='approximate'` indicates that the calculation of the
        permutation test will be approximate, i.e., not all possible permutations
        will be generated.
        Instead, random permutations of the attributes to test will be generated.
        On the other hand, the argument `p_value_iterations`
        indicates the number of permutations that will be generated and tested.

        >>> WEAT().run_query(
        ...     query,
        ...     model,
        ...     calculate_p_value=True,
        ...     p_value_method="approximate",
        ...     p_value_iterations=10000,
        ... )
        {
            'query_name': 'Female terms and Male Terms wrt Family and Career',
            'result': 0.46343879750929773,
            'weat': 0.46343879750929773,
            'effect_size': 0.4507652708557911,
            'p_value': 0.1865813418658134
        }

        """
        # check the types of the provided arguments (only the defaults).
        self._check_input(
            query=query,
            model=model,
            lost_vocabulary_threshold=lost_vocabulary_threshold,
            warn_not_found_words=warn_not_found_words,
        )

        # transform query word sets into embeddings
        embeddings = get_embeddings_from_query(
            model=model,
            query=query,
            lost_vocabulary_threshold=lost_vocabulary_threshold,
            preprocessors=preprocessors,
            strategy=strategy,
            normalize=normalize,
            warn_not_found_words=warn_not_found_words,
        )

        # if there is any/some set has less words than the allowed limit,
        # return the default value (nan)
        if embeddings is None:
            return {
                "query_name": query.query_name,
                "result": np.nan,
                "weat": np.nan,
                "effect_size": np.nan,
            }

        # get the targets and attribute sets transformed into embeddings.
        target_sets, attribute_sets = embeddings

        # get only the embeddings of the sets.
        target_embeddings = list(target_sets.values())
        attribute_embeddings = list(attribute_sets.values())

        target_0 = np.array(list(target_embeddings[0].values()))
        target_1 = np.array(list(target_embeddings[1].values()))
        attribute_0 = np.array(list(attribute_embeddings[0].values()))
        attribute_1 = np.array(list(attribute_embeddings[1].values()))

        # if the requested value is the effect size:
        weat_effect_size = self._calc_effect_size(
            target_0, target_1, attribute_0, attribute_1
        )
        weat = self._calc_weat(target_0, target_1, attribute_0, attribute_1)

        if calculate_p_value:
            p_value = self._calc_p_value(
                target_embeddings=target_embeddings,
                attribute_embeddings=attribute_embeddings,
                original_score=weat,
                test_type=p_value_test_type,
                method=p_value_method,
                iterations=p_value_iterations,
                verbose=p_value_verbose,
            )
        else:
            p_value = np.nan

        # return in result field effect_size
        if return_effect_size:
            return {
                "query_name": query.query_name,
                "result": weat_effect_size,
                "weat": weat,
                "effect_size": weat_effect_size,
                "p_value": p_value,
            }

        # return in result field weat
        return {
            "query_name": query.query_name,
            "result": weat,
            "weat": weat,
            "effect_size": weat_effect_size,
            "p_value": p_value,
        }



"""Module that implements the Query object."""

from itertools import combinations
from typing import Any

import numpy as np


class Query:
    """A container for attribute and target word sets."""

    def __init__(
        self,
        target_sets: list[Any],
        attribute_sets: list[Any],
        target_sets_names: list[str] | None = None,
        attribute_sets_names: list[str] | None = None,
    ) -> None:
        """Initializes the container. It could include a name for each word set.

        Parameters
        ----------
        target_sets : Union[np.ndarray, list]
            Array or list that contains the target word sets.
        attribute_sets : Union[np.ndarray, Iterable]
            Array or list that contains the attribute word sets.
        target_sets_names : Union[np.ndarray, Iterable], optional
            Array or list that contains the word sets names, by default None
        attribute_sets_names : Union[np.ndarray, Iterable], optional
            Array or list that contains the attribute sets names,
            by default None

        Attributes
        ----------
        target_sets : list
            Array or list with the lists of target words.
        attribute_sets : list
            Array or list with the lists of target words.
        template : tuple
            A tuple that contains the template: the cardinality of the target
            and attribute sets respectively.
        target_sets_names : list
            Array or list with the names of target sets.
        attribute_sets_names : list
            Array or list with the lists of target words.
        query_name : str
            A string that contains the auto-generated name of the query.

        Raises
        ------
        TypeError
            if target_sets are not an iterable or np.ndarray instance.
        TypeError
            if attribute_sets are not an iterable or np.ndarray instance.
        Exception
            if the length of target_sets is 0.
        TypeError
            if some element of target_sets is not an array or list.
        TypeError
            if some element of some target set is not an string.
        TypeError
            if some element of attribute_sets is not an array or list.
        TypeError
            if some element of some attribute set is not an string.

        Examples
        --------
        Construct a Query with 2 sets of target words and one set of
        attribute words.

        >>> male_terms = ['male', 'man', 'boy']
        >>> female_terms = ['female', 'woman', 'girl']
        >>> science_terms = ['science','technology','physics']
        >>> query = Query([male_terms, female_terms], [science_terms],
        ...               ['Male terms', 'Female terms'], ['Science terms'])
        >>> query.target_sets
        [['male', 'man', 'boy'], ['female', 'woman', 'girl']]
        >>> query.attribute_sets
        [['science', 'technology', 'physics']]
        >>> query.query_name
        'Male terms and Female terms wrt Science terms'

        """
        # check input type
        if not isinstance(target_sets, list | np.ndarray):
            raise TypeError(
                f"target_sets must be a numpy array or list. Given: {type(target_sets)}"
            )

        if not isinstance(attribute_sets, list | np.ndarray):
            raise TypeError(
                f"attribute_sets must be a numpy array or list. "
                f"Given: {type(attribute_sets)}"
            )

        # check input array sizes
        if len(target_sets) == 0:
            raise Exception(
                "target_sets must have at least one array or list of words. "
                f"given: {target_sets}"
            )

        # check all words that target sets contains.
        for idx, target_set in enumerate(target_sets):
            if not isinstance(target_set, np.ndarray | list):
                raise TypeError(
                    "Each target set must be a list or an array of strings. "
                    f"Given: {type(target_set)} at postion {idx}"
                )
            for word_idx, word in enumerate(target_set):
                if not isinstance(word, str):
                    raise TypeError(
                        f"All elements in target set {idx} must be strings. "
                        f"Given: {type(word)} at position {word_idx}"
                    )

        # check all words that attribute sets contains.
        for idx, attribute_set in enumerate(attribute_sets):
            if not isinstance(attribute_set, np.ndarray | list):
                raise TypeError(
                    "Each attribute set must be a list or an array of strings."
                    f" Given: {type(attribute_set)} at postion {idx}"
                )
            for word_idx, word in enumerate(attribute_set):
                if not isinstance(word, str):
                    raise TypeError(
                        f"All elements in attribute set {idx} must be strings. "
                        f"Given: {type(word)} at position {word_idx}"
                    )

        # set target and attributes sets to this instance.
        self.target_sets = target_sets
        self.attribute_sets = attribute_sets

        # set the template/cardinality (t, a) of the sets
        self.template = (len(target_sets), len(attribute_sets))

        # set target sets names.
        if target_sets_names is None:
            self.target_sets_names = [
                f"Target set {i}" for i in range(self.template[0])
            ]
        else:
            if len(target_sets_names) != self.template[0]:
                raise ValueError(
                    f"target_sets (len={self.template[0]}) does not have the same "
                    "number of elements as target_sets_names "
                    f"(len={len(target_sets_names)})"
                )
            self.target_sets_names = target_sets_names

        # set attribute and attribute sets names.
        if attribute_sets_names is None:
            self.attribute_sets_names = [
                f"Attribute set {i}" for i in range(self.template[1])
            ]
        else:
            if len(attribute_sets_names) != self.template[1]:
                raise ValueError(
                    f"attribute_sets (len={self.template[1]}) does not have the same "
                    "number of elements as attribute_sets_names "
                    f"(len={len(attribute_sets_names)})"
                )
            self.attribute_sets_names = attribute_sets_names

        self.query_name = self._get_query_name()

    def __eq__(self, other: Any) -> bool:
        """Check if some object is equal to this query.

        Steps:
        - Check other type.
        - Compare template.
        - Check if the number of target sets in both queries is equal.
        - Check if the number of attribute sets in both queries is equal.
        - Check if every target set is the same in both queries.
        - Check if every attribute set is the same in both queries.
        - Check if the names of the target sets are equal in both queries.
        - Check if the names of the attribute sets are equal in both queries.

        Parameters
        ----------
        other : Any
            The object to compare.

        Returns
        -------
        bool
            True if other is the same query, False in any other case.

        """
        if not isinstance(other, Query):
            return False

        if self.template[0] != other.template[0]:
            return False
        if self.template[1] != other.template[1]:
            return False

        if len(self.target_sets) != len(other.target_sets):
            return False
        if len(self.attribute_sets) != len(other.attribute_sets):
            return False

        for target_set, other_target_set in zip(
            self.target_sets,
            other.target_sets,
            strict=False,
        ):
            if target_set != other_target_set:
                return False

        for attribute_set, other_attribute_set in zip(
            self.attribute_sets,
            other.attribute_sets,
            strict=False,
        ):
            if attribute_set != other_attribute_set:
                return False

        for names, other_names in zip(
            self.target_sets_names,
            other.target_sets_names,
            strict=False,
        ):
            if names != other_names:
                return False

        for names, other_names in zip(
            self.attribute_sets_names,
            other.attribute_sets_names,
            strict=False,
        ):
            if names != other_names:
                return False
        return True

    def __repr__(self) -> str:
        """Generates a repr that shows the name, target and attributes of the query.

        Returns
        -------
        str
            The generated representation.

        """
        try:
            repr_ = (
                "<Query: "
                + self.query_name
                + "\n- Target sets: "
                + repr(self.target_sets)
                + "\n- Attribute sets:"
                + repr(self.attribute_sets)
                + ">"
            )
            return repr_
        except AttributeError:
            # it can happen if some of the attributes (query_name, target_sets
            # or attribute_sets) are not defined.
            return "<Query with wrong __repr__>"

    def dict(self) -> dict[str, Any]:
        """Generate a dictionary from the Query data.

        This includes the target and attribute sets, as well as their names,
        the query name generated from them and the query template.

        Returns
        -------
        Dict[str, Any]
            The dictionary generated with the query data.

        """
        return {
            "target_sets": self.target_sets,
            "attribute_sets": self.attribute_sets,
            "target_sets_names": self.target_sets_names,
            "attribute_sets_names": self.attribute_sets_names,
            "query_name": self.query_name,
            "template": self.template,
        }

    def get_subqueries(self, new_template: tuple) -> list:
        """Generate the subqueries from this query using the given template."""
        if not isinstance(new_template[0], int):
            raise TypeError(
                "The new target cardinality (new_template[0]) must be int. "
                f"Given: {new_template[0]}"
            )
        if not isinstance(new_template[1], int):
            raise TypeError(
                "The new attribute cardinality (new_template[1]) must be int. "
                f"Given: {new_template[1]}"
            )

        if new_template[0] > self.template[0]:
            raise Exception(
                "The new target cardinality (new_template[0]) must be equal or "
                f"less than the original target set cardinality. "
                f"Given: {new_template[0]}"
            )
        if new_template[1] > self.template[1]:
            raise Exception(
                "The new attribute cardinality (new_template[1]) must be equal"
                " or less than the original attribute set cardinality. "
                f"Given: {new_template[1]}"
            )

        target_combinations = list(
            combinations(range(self.template[0]), new_template[0])
        )
        attribute_combinations = list(
            combinations(range(self.template[1]), new_template[1])
        )

        target_subsets = [
            [self.target_sets[idx] for idx in combination]
            for combination in target_combinations
        ]
        target_subsets_names = [
            [self.target_sets_names[idx] for idx in combination]
            for combination in target_combinations
        ]
        attribute_subsets = [
            [self.attribute_sets[idx] for idx in combination]
            for combination in attribute_combinations
        ]
        attribute_subsets_names = [
            [self.attribute_sets_names[idx] for idx in combination]
            for combination in attribute_combinations
        ]

        subqueries = [
            [
                Query(
                    target_subset,
                    attribute_subset,
                    target_subset_name,
                    attribute_subset_name,
                )
                for attribute_subset, attribute_subset_name in zip(
                    attribute_subsets, attribute_subsets_names, strict=False
                )
            ]
            for target_subset, target_subset_name in zip(
                target_subsets, target_subsets_names, strict=False
            )
        ]

        return np.array(subqueries).flatten().tolist()

    def _get_query_name(self) -> str:
        """Generate the query name from the name of its target and attribute sets.

        Returns
        -------
        str
            The name of the query.

        """
        target_sets_names = self.target_sets_names
        attribute_sets_names = self.attribute_sets_names

        if len(target_sets_names) == 1:
            target = target_sets_names[0]
        elif len(target_sets_names) == 2:
            target = target_sets_names[0] + " and " + target_sets_names[1]
        else:
            target = (
                ", ".join([str(x) for x in target_sets_names[0:-1]])
                + " and "
                + target_sets_names[-1]
            )

        if len(attribute_sets_names) == 0:
            return target

        if len(attribute_sets_names) == 1:
            attribute = attribute_sets_names[0]
        elif len(attribute_sets_names) == 2:
            attribute = attribute_sets_names[0] + " and " + attribute_sets_names[1]
        else:
            attribute = (
                ", ".join([str(x) for x in attribute_sets_names[0:-1]])
                + " and "
                + attribute_sets_names[-1]
            )

        return target + " wrt " + attribute



"""Module with utilities that ease the transformation of word sets to embeddings."""

from collections.abc import Callable, Sequence
import logging

import numpy as np
from sklearn.feature_extraction.text import strip_accents_ascii, strip_accents_unicode

from wefe.query import Query
from wefe.word_embedding_model import WordEmbeddingModel

EmbeddingDict = dict[str, np.ndarray]
EmbeddingSets = dict[str, EmbeddingDict]


def preprocess_word(
    word: str,
    options: dict[str, str | bool | Callable] = {},
    vocab_prefix: str | None = None,
) -> str:
    """pre-processes a word before it is searched in the model's vocabulary.

    Parameters
    ----------
    word : str
        Word to be preprocessed.
    options : Dict[str, Union[str, bool, Callable]], optional
        Dictionary with arguments that specifies how the words will be preprocessed,
        The available word preprocessing options are as follows:

        - ```lowercase```: bool. Indicates if the words are transformed to lowercase.

        - ```uppercase```: bool. Indicates if the words are transformed to uppercase.

        - ```titlecase```: bool. Indicates if the words are transformed to titlecase.

        - ```strip_accents```: `bool`, `{'ascii', 'unicode'}`: Specifies if the
          accents of the words are eliminated. The stripping type can be
          specified. True uses 'unicode' by default.

        - ```preprocessor```: Callable. It receives a function that operates on each
          word. In the case of specifying a function, it overrides
          the default preprocessor (i.e., the previous options
          stop working).

        By default, no preprocessing is generated, which is equivalent to {}

    Returns
    -------
    str
        The pre-processed word according to the given parameters.

    """
    preprocessor = options.get("preprocessor")

    # if the preprocessor is specified, it takes precedence over all other operations.
    if preprocessor is not None and callable(preprocessor):
        word = preprocessor(word)

    else:
        strip_accents = options.get("strip_accents", False)
        lowercase = options.get("lowercase", False)
        uppercase = options.get("uppercase", False)
        titlecase = options.get("titlecase", False)

        if lowercase:
            word = word.lower()
        elif uppercase:
            word = word.upper()
        elif titlecase:
            word = word.title()

        # by default, if strip_accents is True, run strip_accents_unicode
        if strip_accents is True:
            word = strip_accents_unicode(word)
        elif strip_accents == "ascii":
            word = strip_accents_ascii(word)
        elif strip_accents == "unicode":
            word = strip_accents_unicode(word)

    if vocab_prefix is not None and isinstance(vocab_prefix, str):
        return vocab_prefix + word
    return word


def get_embeddings_from_set(
    model: WordEmbeddingModel,
    word_set: Sequence[str],
    preprocessors: list[dict[str, str | bool | Callable]] = [{}],
    strategy: str = "first",
    normalize: bool = False,
    verbose: bool = False,
) -> tuple[list[str], dict[str, np.ndarray]]:
    """Transform a sequence of words into dictionary that maps word - word embedding.

    The method discard out words that are not in the model's vocabulary
    (according to the rules specified in the preprocessors).


    Parameters
    ----------
    model : WordEmbeddingModel
        A word embeddding model

    word_set : Sequence[str]
        A sequence with the words that this function will convert to embeddings.

    preprocessors : List[Dict[str, Union[str, bool, Callable]]]
        A list with preprocessor options.

        A ``preprocessor`` is a dictionary that specifies what processing(s) are
        performed on each word before it is looked up in the model vocabulary.
        For example, the ``preprocessor``
        ``{'lowecase': True, 'strip_accents': True}`` allows you to lowercase
        and remove the accent from each word before searching for them in the
        model vocabulary. Note that an empty dictionary ``{}`` indicates that no
        preprocessing is done.

        The possible options for a preprocessor are:

        *   ``lowercase``: ``bool``. Indicates that the words are transformed to
            lowercase.
        *   ``uppercase``: ``bool``. Indicates that the words are transformed to
            uppercase.
        *   ``titlecase``: ``bool``. Indicates that the words are transformed to
            titlecase.
        *   ``strip_accents``: ``bool``, ``{'ascii', 'unicode'}``: Specifies that
            the accents of the words are eliminated. The stripping type can be
            specified. True uses 'unicode' by default.
        *   ``preprocessor``: ``Callable``. It receives a function that operates
            on each word. In the case of specifying a function, it overrides the
            default preprocessor (i.e., the previous options stop working).

        A list of preprocessor options allows you to search for several
        variants of the words into the model. For example, the preprocessors
        ``[{}, {"lowercase": True, "strip_accents": True}]``
        ``{}`` allows searching first for the original words in the vocabulary of
        the model. In case some of them are not found,
        ``{"lowercase": True, "strip_accents": True}`` is executed on these words
        and then they are searched in the model vocabulary.
        by default [{}]

    strategy : str, optional
        The strategy indicates how it will use the preprocessed words: 'first' will
        include only the first transformed word found. 'all' will include all
        transformed words found, by default "first".

    normalize : bool, optional
        True indicates that embeddings will be normalized, by default False

    verbose : bool, optional
        Indicates whether the execution status of this function is printed,
        by default False

    Returns
    -------
    Tuple[List[str], Dict[str, np.ndarray]]
        A tuple containing the words that could not be found and a dictionary with
        the found words and their corresponding embeddings.

    """
    # ----------------------------------------------------------------------------------
    # type verifications.

    if not isinstance(model, WordEmbeddingModel):
        raise TypeError(f"model should be a WordEmbeddingModel instance, got {model}.")

    if not isinstance(word_set, list | tuple | np.ndarray):
        raise TypeError(
            f"word_set should be a list, tuple or np.array of strings, got {word_set}."
        )

    if not isinstance(preprocessors, list):
        raise TypeError(
            "preprocessors should be a list of dicts which contains preprocessor "
            f"options, got {preprocessors}."
        )
    if len(preprocessors) == 0:
        raise TypeError(
            "preprocessors must indicate at least one preprocessor, even if it is "
            "an empty dictionary {}, "
            f"got: {preprocessors}."
        )
    for idx, p in enumerate(preprocessors):
        if not isinstance(p, dict):
            raise TypeError(
                f"each preprocessor should be a dict, got {p} at index {idx}."
            )

    if strategy != "first" and strategy != "all":
        raise ValueError(f"strategy should be 'first' or 'all', got {strategy}.")

    # ----------------------------------------------------------------------------------
    # filter the words

    selected_embeddings = {}
    not_found_words = []

    for word in word_set:
        for preprocessor in preprocessors:
            preprocessed_word = preprocess_word(
                word, options=preprocessor, vocab_prefix=model.vocab_prefix
            )
            embedding = model[preprocessed_word] if preprocessed_word in model else None  # noqa: SIM401

            if embedding is not None:
                selected_embeddings[preprocessed_word] = embedding

                # if the selected strategy is first, then it stops on the first
                # word encountered.
                if strategy == "first":
                    break
            else:
                not_found_words.append(preprocessed_word)

    # if requested, normalize embeddings.
    if normalize:
        selected_embeddings = {
            k: v / np.linalg.norm(v) for k, v in selected_embeddings.items()
        }

    if verbose:
        print(
            f"Word(s) found: {list(selected_embeddings.keys())}, "
            f"not found: {not_found_words}"
        )

    return not_found_words, selected_embeddings


def _warn_not_found_words(
    warn_not_found_words: bool,
    not_found_words: list[str],
    model_name: str,
    set_name: str,
) -> None:
    if not isinstance(warn_not_found_words, bool):
        raise TypeError(
            f"warn_not_found_words should be a boolean, got {warn_not_found_words}."
        )

    if warn_not_found_words and len(not_found_words) > 0:
        logging.warning(
            f"The following words from set '{set_name}' do not exist within the "
            f"vocabulary of {model_name}: {not_found_words}"
        )


def _check_lost_vocabulary_threshold(
    model: WordEmbeddingModel,
    embeddings: EmbeddingDict,
    word_set: list[str],
    word_set_name: str,
    lost_vocabulary_threshold: float,
) -> None:
    if not isinstance(lost_vocabulary_threshold, float | np.floating):
        raise TypeError(
            "lost_vocabulary_threshold should be float, "
            f"got {lost_vocabulary_threshold}."
        )

    remaining_words = list(embeddings.keys())
    number_of_lost_words = len(word_set) - len(remaining_words)
    percentage_of_lost_words = number_of_lost_words / len(word_set)

    # if the percentage of filtered words are greater than the
    # threshold, log and return False
    if percentage_of_lost_words > lost_vocabulary_threshold:
        logging.warning(
            f"The transformation of '{word_set_name}' into {model.name} embeddings "
            f"lost proportionally more words than specified in 'lost_words_threshold': "
            f"{round(percentage_of_lost_words, 2)} lost with respect to "
            f"{lost_vocabulary_threshold} maximum loss allowed."
        )
        return True
    return False


def get_embeddings_from_tuples(
    model: WordEmbeddingModel,
    sets: Sequence[Sequence[str]],
    sets_name: str | None = None,
    preprocessors: list[dict[str, str | bool | Callable]] = [{}],
    strategy: str = "first",
    normalize: bool = False,
    discard_incomplete_sets: bool = True,
    warn_lost_sets: bool = True,
    verbose: bool = False,
) -> list[EmbeddingDict]:
    """Given a sequence of word sets, obtain their corresponding embeddings.

    Parameters
    ----------
    model

    sets :  Sequence[Sequence[str]]
        A sequence containing word sets.
        Example: `[['woman', 'man'], ['she', 'he'], ['mother', 'father'] ...]`.

    sets_name : Union[str, optional]
        The name of the set of word sets. Example: `definning sets`.
        This parameter is used only for printing.
        by default None

    preprocessors : List[Dict[str, Union[str, bool, Callable]]]
        A list with preprocessor options.

        A ``preprocessor`` is a dictionary that specifies what processing(s) are
        performed on each word before it is looked up in the model vocabulary.
        For example, the ``preprocessor``
        ``{'lowecase': True, 'strip_accents': True}`` allows you to lowercase
        and remove the accent from each word before searching for them in the
        model vocabulary. Note that an empty dictionary ``{}`` indicates that no
        preprocessing is done.

        The possible options for a preprocessor are:

        *   ``lowercase``: ``bool``. Indicates that the words are transformed to
            lowercase.
        *   ``uppercase``: ``bool``. Indicates that the words are transformed to
            uppercase.
        *   ``titlecase``: ``bool``. Indicates that the words are transformed to
            titlecase.
        *   ``strip_accents``: ``bool``, ``{'ascii', 'unicode'}``: Specifies that
            the accents of the words are eliminated. The stripping type can be
            specified. True uses 'unicode' by default.
        *   ``preprocessor``: ``Callable``. It receives a function that operates
            on each word. In the case of specifying a function, it overrides the
            default preprocessor (i.e., the previous options stop working).

        A list of preprocessor options allows you to search for several
        variants of the words into the model. For example, the preprocessors
        ``[{}, {"lowercase": True, "strip_accents": True}]``
        ``{}`` allows searching first for the original words in the vocabulary of
        the model. In case some of them are not found,
        ``{"lowercase": True, "strip_accents": True}`` is executed on these words
        and then they are searched in the model vocabulary.
        by default [{}]

    strategy : str, optional
        The strategy indicates how it will use the preprocessed words: 'first' will
        include only the first transformed word found. 'all' will include all
        transformed words found, by default "first".

    normalize : bool, optional
        True indicates that embeddings will be normalized, by default False

    discard_incomplete_sets : bool, optional
        True indicates that if a set could not be completely converted, it will be
        discarded., by default True

    warn_lost_sets : bool, optional
        Indicates whether word sets that cannot be fully converted to embeddings
        are warned in the logger,
        by default True

    verbose : bool, optional
        Indicates whether the execution status of this function is printed,
        by default False

    Returns
    -------
    List[EmbeddingDict]
        A list of dictionaries. Each dictionary contains as keys a pair of words
        and as values their associated embeddings.

    """
    if not isinstance(sets, list | tuple | np.ndarray):
        raise TypeError(
            "sets should be a sequence of sequences (list, tuple or np.array) "
            f"of strings, got: {type(sets)}."
        )

    for idx, set_ in enumerate(sets):
        if not isinstance(set_, list | tuple | np.ndarray):
            raise TypeError(
                "Every set in sets should be a list, tuple or np.array of "
                f"strings, got in index {idx}: {type(set_)}"
            )

        for word_idx, word in enumerate(set_):
            if not isinstance(word, str):
                raise TypeError(
                    "All set elements in a set of words should be strings. "
                    f"Got in set {idx} at position {word_idx}: {type(word)}"
                )

    if sets_name is not None and not isinstance(sets_name, str):
        raise TypeError(f"sets_name should be a string or None, got: {type(sets_name)}")
    if not isinstance(warn_lost_sets, bool):
        raise TypeError(f"warn_lost_sets should be a bool, got: {type(warn_lost_sets)}")
    if not isinstance(verbose, bool):
        raise TypeError(f"verbose should be a bool, got: {type(verbose)}")

    embedding_sets: list[EmbeddingDict] = []

    # For each definitional pair:
    for set_idx, set_ in enumerate(sets):
        # Transform the pair to a embedding dict.
        # idea: (word_1, word_2) -> {'word_1': embedding, 'word_2'.: embedding}
        # TODO: Add identifier of the set that is being transformed.
        # if verbose:
        #     print(f"Transforming '{}' set ")
        not_found_words, embedding_pair = get_embeddings_from_set(
            model, set_, preprocessors, strategy, normalize, verbose
        )

        # If some word of the current pair can not be converted, discard the pair.
        if discard_incomplete_sets and len(not_found_words) > 0 and warn_lost_sets:
            set_name = f" of {sets_name} pair" if sets_name else ""
            logging.warning(
                f"Word(s) {not_found_words}{set_name} at index {set_idx} "
                "were not found. This pair will be omitted."
            )
        else:
            if normalize:
                for word in embedding_pair:
                    embedding = embedding_pair[word]
                    normalized_embedding = embedding / np.linalg.norm(embedding)
                    if np.linalg.norm(embedding) < 1:
                        normalized_embedding = embedding / np.linalg.norm(embedding)
                    embedding_pair[word] = normalized_embedding

            embedding_sets.append(embedding_pair)

    if len(embedding_sets) == 0:
        set_name = f"from the set {sets_name} " if sets_name else ""
        msg = (
            f"No set {set_name}could be converted to embedding because no set "
            "could be fully found in the model vocabulary."
        )
        raise Exception(msg)

    if verbose:
        print(
            f"{len(embedding_sets)}/{len(sets)} sets of "
            "words were correctly converted to sets of embeddings"
        )

    return embedding_sets


def get_embeddings_from_query(
    model: WordEmbeddingModel,
    query: Query,
    lost_vocabulary_threshold: float = 0.2,
    preprocessors: list[dict[str, str | bool | Callable]] = [{}],
    strategy: str = "first",
    normalize: bool = False,
    warn_not_found_words: bool = False,
    verbose: bool = False,
) -> tuple[EmbeddingSets, EmbeddingSets] | None:
    """Obtain the word vectors associated with the provided Query.

    The words that does not appears in the word embedding pretrained model
    vocabulary under the specified pre-processing are discarded.
    If the remaining words percentage in any query set is lower than the specified
    threshold, the function will return None.


    Parameters
    ----------
    query : Query
        The query to be processed.

    lost_vocabulary_threshold : float, optional, by default 0.2
        Indicates the proportional limit of words that any set of the query is
        allowed to lose when transforming its words into embeddings.
        In the case that any set of the query loses proportionally more words
        than this limit, this method will return None.

    preprocessors : List[Dict[str, Union[str, bool, Callable]]]
        A list with preprocessor options.

        A ``preprocessor`` is a dictionary that specifies what processing(s) are
        performed on each word before it is looked up in the model vocabulary.
        For example, the ``preprocessor``
        ``{'lowecase': True, 'strip_accents': True}`` allows you to lowercase
        and remove the accent from each word before searching for them in the
        model vocabulary. Note that an empty dictionary ``{}`` indicates that no
        preprocessing is done.

        The possible options for a preprocessor are:

        *   ``lowercase``: ``bool``. Indicates that the words are transformed to
            lowercase.
        *   ``uppercase``: ``bool``. Indicates that the words are transformed to
            uppercase.
        *   ``titlecase``: ``bool``. Indicates that the words are transformed to
            titlecase.
        *   ``strip_accents``: ``bool``, ``{'ascii', 'unicode'}``: Specifies that
            the accents of the words are eliminated. The stripping type can be
            specified. True uses 'unicode' by default.
        *   ``preprocessor``: ``Callable``. It receives a function that operates
            on each word. In the case of specifying a function, it overrides the
            default preprocessor (i.e., the previous options stop working).

        A list of preprocessor options allows you to search for several
        variants of the words into the model. For example, the preprocessors
        ``[{}, {"lowercase": True, "strip_accents": True}]``
        ``{}`` allows searching first for the original words in the vocabulary of
        the model. In case some of them are not found,
        ``{"lowercase": True, "strip_accents": True}`` is executed on these words
        and then they are searched in the model vocabulary.
        by default [{}]

    strategy : str, optional
        The strategy indicates how it will use the preprocessed words: 'first' will
        include only the first transformed word found. 'all' will include all
        transformed words found, by default "first".

    normalize : bool, optional
        True indicates that embeddings will be normalized, by default False

    warn_not_found_words : bool, optional
        A flag that indicates if the function will warn (in the logger)
        the words that were not found in the model's vocabulary,
        by default False.

    verbose : bool, optional
        Indicates whether the execution status of this function is printed,
        by default False

    Returns
    -------
    Union[Tuple[EmbeddingSets, EmbeddingSets], None]
        A tuple of dictionaries containing the targets and attribute sets or None
        in case there is a set that has proportionally less embeddings than it was
        allowed to lose.

    """
    # Type check
    if not isinstance(query, Query):
        raise TypeError(f"query should be an instance of Query, got {query}.")

    some_set_lost_more_words_than_threshold: bool = False

    target_embeddings: EmbeddingSets = {}
    attribute_embeddings: EmbeddingSets = {}

    # --------------------------------------------------------------------
    # get target sets embeddings
    for target_set, target_set_name in zip(
        query.target_sets, query.target_sets_names, strict=False
    ):
        not_found_words, obtained_embeddings = get_embeddings_from_set(
            model=model,
            word_set=target_set,
            preprocessors=preprocessors,
            strategy=strategy,
            normalize=normalize,
            verbose=verbose,
        )

        # warn not found words if it is enabled.
        _warn_not_found_words(
            warn_not_found_words, not_found_words, model.name, target_set_name
        )

        # if the lost words are greater than the threshold,
        # warn and change the flag.
        if _check_lost_vocabulary_threshold(
            model,
            obtained_embeddings,
            target_set,
            target_set_name,
            lost_vocabulary_threshold,
        ):
            some_set_lost_more_words_than_threshold = True

        target_embeddings[target_set_name] = obtained_embeddings

    # --------------------------------------------------------------------
    # get attribute sets embeddings
    for attribute_set, attribute_set_name in zip(
        query.attribute_sets, query.attribute_sets_names, strict=False
    ):
        not_found_words, obtained_embeddings = get_embeddings_from_set(
            model=model,
            word_set=attribute_set,
            preprocessors=preprocessors,
            strategy=strategy,
            normalize=normalize,
            verbose=verbose,
        )

        _warn_not_found_words(
            warn_not_found_words, not_found_words, model.name, attribute_set_name
        )

        # if the filtered words are greater than the threshold,
        # log and change the flag.
        if _check_lost_vocabulary_threshold(
            model,
            obtained_embeddings,
            attribute_set,
            attribute_set_name,
            lost_vocabulary_threshold,
        ):
            some_set_lost_more_words_than_threshold = True

        attribute_embeddings[attribute_set_name] = obtained_embeddings

    # check if some set has fewer words than the threshold. if that's
    #  the case, return None
    if some_set_lost_more_words_than_threshold:
        logging.error(
            f"At least one set of '{query.query_name}' query has proportionally fewer "
            f"embeddings than allowed by the lost_vocabulary_threshold parameter "
            f"({lost_vocabulary_threshold}). This query will return np.nan."
        )
        return None

    return target_embeddings, attribute_embeddings



"""Relative Negative Sentiment Bias (RNSB) metric implementation."""

from collections.abc import Callable
import logging
from typing import Any

import numpy as np
import pandas as pd
from scipy.stats import entropy
from sklearn.base import BaseEstimator
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import classification_report
from sklearn.model_selection import train_test_split

from wefe.metrics.base_metric import BaseMetric
from wefe.preprocessing import get_embeddings_from_query
from wefe.query import Query
from wefe.word_embedding_model import WordEmbeddingModel


class RNSB(BaseMetric):
    r"""Relative Relative Negative Sentiment Bias (RNSB).

    The metric was originally proposed in "A transparent framework for evaluating
    unintended demographic bias in word embeddings" [1].

    This metric is based on measuring bias through word sentiment.
    The main idea is that if there was no bias, all words should be equally negative.
    Therefore, its procedure is based on calculating how negative the words in
    the target sets are.

    For this purpose, RNSB trains a classifier that assigns a probability to each
    word of belonging to the negative class (in the original work the classifier is
    trained using
    :func:`~wefe.datasets.load_bingliu`
    of positive and negative words).
    Then, it generates a probability distribution with the probabilities calculated in
    the previous step and compares them to the uniform distribution
    (case where all words have the same probability of being negative) using
    KL divergence.

    When the negative probability distribution is equal to the uniform one (i.e., there
    is no bias), the KL divergence is 0.

    The following description of the metric is WEFE's adaptation of what was presented
    in the original RNSB work.

    RNSB receives as input queries with two attribute sets :math:`A_1` and
    :math:`A_2` and two or more target sets, thus has a template (tuple of numbers that
    defines the allowed target and attribute sets in the query)
    of the form :math:`s=(N,2)` with :math:`N\geq 2`.

    Given a query :math:`Q=(\{T_1,T_2,\ldots,T_n\},\{A_1,A_2\})` RNSB is
    calculated under the following steps:

    1. First constructs a binary classifier  :math:`C_{(A_1,A_2)}(\cdot)` using
       set :math:`A_1` as training examples for the negative class, and :math:`A_2` as
       training examples for the positive class.

    2. After the training process, this classifier gives for every word :math:`w` a
       probability :math:`C_{(A_1,A_2)}(w)` that can be interpreted as the degree of
       association of :math:`w` with respect to  :math:`A_2` (value
       :math:`1-C_{(A_1,A_2)}(w)` is the degree of association with :math:`A_1`).

    3. Then, the metric constructs a probability distribution :math:`P(\cdot)` over all
       the words :math:`w` in :math:`T_1\cup \cdots \cup T_n`, by computing
       :math:`C_{(A_1,A_2)}(w)` and normalizing it to ensure that
       :math:`\sum_w P(w)=1`.

    4. Finally RNSB is calculated as the distance between :math:`P(\cdot)` and
       the uniform distribution :math:`Y(\cdot)` using the KL-divergence.

    The main idea behind RNSB is that the more that :math:`P(\cdot)` resembles a
    uniform distribution, the less biased the word embedding model is.
    Thus, the optimal value is 0.

    You can see the full paper replication in Previous Studies Replication section.

    References
    ----------
    | [1]: Chris Sweeney and Maryam Najafian. A transparent framework for evaluating
           unintended demographic bias in word embeddings.
    |      In Proceedings of the 57th Annual Meeting of the Association for
           Computational Linguistics, pages 1662–1667, 2019.
    | [2]: https://github.com/ChristopherSweeney/AIFairness/blob/master/python_notebooks/Measuring_and_Mitigating_Word_Embedding_Bias.ipynb

    """  # noqa: E501

    metric_template = ("n", 2)
    metric_name = "Relative Negative Sentiment Bias"
    metric_short_name = "RNSB"

    def _train_classifier(
        self,
        attribute_embeddings_dict: list[dict[str, np.ndarray]],
        estimator: BaseEstimator,
        estimator_params: dict[str, Any],
        random_state: int | None,
        holdout: bool,
        print_model_evaluation: bool,
    ) -> tuple[BaseEstimator, float]:
        """Train the sentiment classifier from the provided attribute embeddings.

        Parameters
        ----------
        attribute_embeddings_dict : dict[str, np.ndarray]
            A dict with the attributes keys and embeddings

        estimator : BaseEstimator, optional
            A scikit-learn classifier class that implements predict_proba function,
            by default LogisticRegression,

        estimator_params : dict, optional
            Parameters that will use the classifier, by default { 'max_iter': 10000, }

        random_state : Union[int, None], optional
            A seed that allows making the execution of the query reproducible, by
            default None

        holdout: bool, optional
            True indicates that a holdout (split attributes in train/test sets) will
            be executed before running the model training.
            This option allows for evaluating the performance of the classifier
            (can be printed using print_model_evaluation=True) at the cost of training
            the classifier with fewer examples. False disables this functionality.
            Note that holdout divides into 80% train and 20% test, performs a shuffle
            and tries to maintain the original ratio of the classes via stratify param,
            by default True.

        print_model_evaluation : bool, optional
            Indicates whether the classifier evaluation is printed after the
            training process is completed, by default False

        Returns
        -------
        Tuple[BaseEstimator, float]
            The trained classifier and the accuracy obtained by the model.

        """
        # when random_state is not none, set it on classifier params.
        if random_state is not None:
            estimator_params["random_state"] = random_state

        # the attribute 0 words are treated as positive words.
        positive_embeddings = np.array(list(attribute_embeddings_dict[0].values()))
        # the attribute 1 words are treated as negative words.
        negative_embeddings = np.array(list(attribute_embeddings_dict[1].values()))

        # generate the labels (1, -1) for each embedding
        positive_labels = np.ones(positive_embeddings.shape[0])
        negative_labels = -np.ones(negative_embeddings.shape[0])

        attributes_embeddings = np.concatenate(
            (positive_embeddings, negative_embeddings)
        )
        attributes_labels = np.concatenate((positive_labels, negative_labels))

        if holdout:
            split = train_test_split(
                attributes_embeddings,
                attributes_labels,
                shuffle=True,
                random_state=random_state,
                test_size=0.2,
                stratify=attributes_labels,
            )
            X_embeddings_train, X_embeddings_test, y_train, y_test = split

            num_train_negative_examples = np.count_nonzero(y_train == -1)
            num_train_positive_examples = np.count_nonzero(y_train == 1)

            # Check the number of train and test examples.
            if num_train_positive_examples == 1:
                raise Exception(
                    "After splitting the dataset using train_test_split "
                    "(with test_size=0.1), the first attribute remained with 0 "
                    "training examples."
                )

            if num_train_negative_examples < 1:
                raise Exception(
                    "After splitting the dataset using train_test_split "
                    "(with test_size=0.1), the second attribute remained with 0 "
                    "training examples."
                )

            estimator = estimator(**estimator_params)
            estimator.fit(X_embeddings_train, y_train)

            # evaluate
            y_pred = estimator.predict(X_embeddings_test)
            score = estimator.score(X_embeddings_test, y_test)

            if print_model_evaluation:
                report = classification_report(
                    y_test, y_pred, labels=estimator.classes_
                )
                print(f"Classification Report:\n{report}")
        else:
            estimator = estimator(**estimator_params)
            estimator.fit(attributes_embeddings, attributes_labels)
            score = estimator.score(attributes_embeddings, attributes_labels)
            if print_model_evaluation:
                print("Holdout is disabled. No evaluation was performed.")

        return estimator, score

    def _calc_rnsb(
        self,
        target_embeddings_dict: list[dict[str, np.ndarray]],
        classifier: BaseEstimator,
    ) -> tuple[np.float64, dict]:
        """Calculate the RNSB metric.

        Parameters
        ----------
        target_embeddings_dict : Dict[str, np.ndarray]
            dict with the target words and their embeddings.
        classifier : BaseEstimator
            Trained scikit-learn classifier in the previous step.

        Returns
        -------
        Tuple[np.float_, dict]
            return the calculated kl_divergence and
            negative_sentiment_probabilities in that order.

        """
        # join the embeddings and the word sets in their respective arrays
        target_embeddings_sets = [
            list(target_dict.values()) for target_dict in target_embeddings_dict
        ]
        target_words_sets = [
            list(target_dict.keys()) for target_dict in target_embeddings_dict
        ]

        # get the probabilities associated with each target word vector
        probabilities = [
            classifier.predict_proba(target_embeddings)
            for target_embeddings in target_embeddings_sets
        ]
        # row where the negative probabilities are located
        negative_class_clf_row = np.where(classifier.classes_ == -1)[0][0]

        # extract only the negative sentiment probability for each word
        negative_probabilities = np.concatenate(
            [
                probability[:, negative_class_clf_row].flatten()
                for probability in probabilities
            ]
        )

        # normalization of the probabilities
        normalized_negative_probabilities = np.array(
            negative_probabilities / np.sum(negative_probabilities)
        )

        # get the uniform dist
        uniform_dist = (
            np.ones(normalized_negative_probabilities.shape[0])
            * 1
            / normalized_negative_probabilities.shape[0]
        )

        # calc the kl divergence
        kl_divergence = entropy(normalized_negative_probabilities, uniform_dist)

        flatten_target_words = [
            item for sublist in target_words_sets for item in sublist
        ]

        # set the probabilities for each word in a dict.
        negative_sentiment_probabilities = dict(
            zip(flatten_target_words, negative_probabilities, strict=False)
        )

        return kl_divergence, negative_sentiment_probabilities

    def run_query(
        self,
        query: Query,
        model: WordEmbeddingModel,
        estimator: BaseEstimator = LogisticRegression,
        estimator_params: dict[str, Any] = {"solver": "liblinear", "max_iter": 10000},
        n_iterations: int = 1,
        random_state: int | None = None,
        holdout: bool = True,
        print_model_evaluation: bool = False,
        lost_vocabulary_threshold: float = 0.2,
        preprocessors: list[dict[str, str | bool | Callable]] = [{}],
        strategy: str = "first",
        normalize: bool = False,
        warn_not_found_words: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Calculate the RNSB metric over the provided parameters.

        Note if you want to use with Bing Liu dataset, you have to pass
        the positive and negative words in the first and second place of
        attribute set array respectively.
        Scores on this metric vary with each run due to different instances
        of classifier training. For this reason, the robustness of these scores
        can be improved by repeating the test several times and returning the
        average of the scores obtained. This can be indicated in the
        n_iterations parameter.

        Parameters
        ----------
        query : Query
            A Query object that contains the target and attribute word sets to
            be tested.

        model : WordEmbeddingModel
            A word embedding model.

        estimator : BaseEstimator
            A scikit-learn classifier class that implements predict_proba function,
            by default None,

        estimator_params : dict
            Parameters that will use the classifier, by default { 'solver': 'liblinear',
            'max_iter': 10000, }

        n_iterations : int, optional
            When provided, it tells the metric to run the specified number of times
            and then average its results. This functionality is indicated to
            strengthen the results obtained.
            Note that you cannot specify random_state next to n_iterations as this
            would always produce the same results,
            by default 1.

        random_state : Union[int, None], optional
            Seed that allows making the execution of the query reproducible.
            Warning: if a random_state other than None is provided along with
            n_iterations, each iteration will split the dataset and train a
            classifier associated to the same seed, so the results of each iteration
            will always be the same, by default None.

        holdout: bool, optional
            True indicates that a holdout (split attributes in train/test sets) will
            be executed before running the model training.
            This option allows for evaluating the performance of the classifier
            (can be printed using print_model_evaluation=True) at the cost of training
            the classifier with fewer examples. False disables this functionality.
            Note that holdout divides into 80% train and 20% test, performs a shuffle
            and tries to maintain the original ratio of the classes via stratify param,
            by default True.

        print_model_evaluation : bool, optional
            Indicates whether the classifier evaluation is printed after the
            training process is completed, by default False

        preprocessors : List[Dict[str, Union[str, bool, Callable]]]
            A list with preprocessor options.

            A ``preprocessor`` is a dictionary that specifies what processing(s) are
            performed on each word before it is looked up in the model vocabulary.
            For example, the ``preprocessor``
            ``{'lowecase': True, 'strip_accents': True}`` allows you to lowercase
            and remove the accent from each word before searching for them in the
            model vocabulary. Note that an empty dictionary ``{}`` indicates that no
            preprocessing is done.

            The possible options for a preprocessor are:

            *   ``lowercase``: ``bool``. Indicates that the words are transformed to
                lowercase.
            *   ``uppercase``: ``bool``. Indicates that the words are transformed to
                uppercase.
            *   ``titlecase``: ``bool``. Indicates that the words are transformed to
                titlecase.
            *   ``strip_accents``: ``bool``, ``{'ascii', 'unicode'}``: Specifies that
                the accents of the words are eliminated. The stripping type can be
                specified. True uses 'unicode' by default.
            *   ``preprocessor``: ``Callable``. It receives a function that operates
                on each word. In the case of specifying a function, it overrides the
                default preprocessor (i.e., the previous options stop working).

            A list of preprocessor options allows you to search for several
            variants of the words into the model. For example, the preprocessors
            ``[{}, {"lowercase": True, "strip_accents": True}]``
            ``{}`` allows searching first for the original words in the vocabulary of
            the model. In case some of them are not found,
            ``{"lowercase": True, "strip_accents": True}`` is executed on these words
            and then they are searched in the model vocabulary.

        strategy : str, optional
            The strategy indicates how it will use the preprocessed words: 'first' will
            include only the first transformed word found. 'all' will include all
            transformed words found, by default "first".

        normalize : bool, optional
            True indicates that embeddings will be normalized, by default False

        warn_not_found_words : bool, optional
            Specifies if the function will warn (in the logger)
            the words that were not found in the model's vocabulary, by default False.

        Returns
        -------
        Dict[str, Any]
            A dictionary with the query name, the calculated kl-divergence,
            the negative probabilities for all tested target words and
            the normalized distribution of probabilities.

        Examples
        --------
        The following example shows how to run a query that measures gender
        bias using RNSB.
        Note that by default the RNSB score is returned plus the negative class
        probabilities for each word and its distribution (the above probabilities
        normalized to 1).

        >>> from wefe.query import Query
        >>> from wefe.utils import load_test_model
        >>> from wefe.metrics import RNSB
        >>>
        >>> # define the query
        >>> query = Query(
        ...     target_sets=[
        ...         ["female", "woman", "girl", "sister", "she", "her", "hers",
        ...          "daughter"],
        ...         ["male", "man", "boy", "brother", "he", "him", "his", "son"],
        ...     ],
        ...     attribute_sets=[
        ...         ["home", "parents", "children", "family", "cousins", "marriage",
        ...          "wedding", "relatives",],
        ...         ["executive", "management", "professional", "corporation", "salary",
        ...          "office", "business", "career", ],
        ...     ],
        ...     target_sets_names=["Female terms", "Male Terms"],
        ...     attribute_sets_names=["Family", "Careers"],
        ... )
        >>>
        >>> # load the model (in this case, the test model included in wefe)
        >>> model = load_test_model()
        >>>
        >>> # instance the metric and run the query
        >>> RNSB().run_query(query, model)
        {
            "query_name": "Female terms and Male Terms wrt Family and Careers",
            "result": 0.10769060995617141,
            "rnsb": 0.10769060995617141,
            "negative_sentiment_probabilities": {
                "female": 0.5742192708509877,
                "woman": 0.32330898567978306,
                "girl": 0.17573260129841273,
                "sister": 0.15229835340332343,
                "she": 0.3761328719677399,
                "her": 0.35739995104539557,
                "hers": 0.2911542159275662,
                "daughter": 0.11714195753410628,
                "male": 0.4550779245077232,
                "man": 0.39826729589696475,
                "boy": 0.17445392462199483,
                "brother": 0.16517694979156405,
                "he": 0.5044892468050808,
                "him": 0.45426103796811057,
                "his": 0.5013980699813614,
                "son": 0.1509265229834842,
            },
            "negative_sentiment_distribution": {
                "female": 0.11103664779476699,
                "woman": 0.0625181838962096,
                "girl": 0.033981372529543176,
                "sister": 0.02944989742595416,
                "she": 0.07273272658861042,
                "her": 0.06911034599602082,
                "hers": 0.0563004234950174,
                "daughter": 0.022651713275710483,
                "male": 0.08799831316676666,
                "man": 0.07701285503210042,
                "boy": 0.033734115115920706,
                "brother": 0.031940228635376634,
                "he": 0.09755296914839981,
                "him": 0.08784035200525282,
                "his": 0.09695522899987082,
                "son": 0.029184626894479145,
            },
        }

        If you want to perform a holdout to evaluate (defualt option) the model
        and print the evaluation, use the params ``holdout=True`` and
        ``print_model_evaluation=True``

        >>> RNSB().run_query(
        ...     query,
        ...     model,
        ...     holdout=True,
        ...     print_model_evaluation=True)
        "Classification Report:"
        "              precision    recall  f1-score   support"
        "                                                     "
        "        -1.0       1.00      1.00      1.00         2"
        "         1.0       1.00      1.00      1.00         2"
        "                                                     "
        "    accuracy                           1.00         4"
        "   macro avg       1.00      1.00      1.00         4"
        "weighted avg       1.00      1.00      1.00         4"
        {
            "query_name": "Female terms and Male Terms wrt Family and Careers",
            "result": 0.09400726375514418,
            "rnsb": 0.09400726375514418,
            "negative_sentiment_probabilities": {
                "female": 0.5583010801302075,
                "woman": 0.3159147912504866,
                "girl": 0.20753840501109977,
                "sister": 0.16020059726421976,
                "she": 0.4266765171984158,
                "her": 0.4066467259229203,
                "hers": 0.32435655424005905,
                "daughter": 0.13318012193912765,
                "male": 0.44129601598998147,
                "man": 0.42681869843678866,
                "boy": 0.21830517614567535,
                "brother": 0.2037443178553553,
                "he": 0.5655603842644314,
                "him": 0.512466010254818,
                "his": 0.5689713390373838,
                "son": 0.18364286185769785,
            },
            "negative_sentiment_distribution": {
                "female": 0.09875108690481095,
                "woman": 0.05587832464521873,
                "girl": 0.0367089439708001,
                "sister": 0.02833593497428311,
                "she": 0.07546961904547295,
                "her": 0.07192679290859644,
                "hers": 0.05737148541506475,
                "daughter": 0.023556611770367462,
                "male": 0.0780554843555203,
                "man": 0.07549476775523993,
                "boy": 0.038613347150078775,
                "brother": 0.03603785404499525,
                "he": 0.1000350969111323,
                "him": 0.09064387893111858,
                "his": 0.10063841921015712,
                "son": 0.032482352007143285,
            },
        }


        If you want to disable the holdout, use the param ``holdout=False``.

        >>> # instance the metric and run the query
        >>> RNSB().run_query(
        ...     query,
        ...     model,
        ...     holdout=False,
        ...     print_model_evaluation=True) # doctest: +SKIP
        "Holdout is disabled. No evaluation was performed."
        {
            "query_name": "Female terms and Male Terms wrt Family and Careers",
            "result": 0.12921977967420623,
            "rnsb": 0.12921977967420623,
            "negative_sentiment_probabilities": {
                "female": 0.5815344717945401,
                "woman": 0.28951392462851366,
                "girl": 0.16744925298532254,
                "sister": 0.1365690846140981,
                "she": 0.40677635339222296,
                "her": 0.3861243765483825,
                "hers": 0.2794312043966708,
                "daughter": 0.1035870893754135,
                "male": 0.45492464330345805,
                "man": 0.4143325974603802,
                "boy": 0.18150440132198242,
                "brother": 0.1607078872193466,
                "he": 0.5656269380025241,
                "him": 0.5025663479575841,
                "his": 0.5657745694122852,
                "son": 0.14803033326417403,
            },
            "negative_sentiment_distribution": {
                "female": 0.10881083995606983,
                "woman": 0.05417091306830872,
                "girl": 0.03133140811261611,
                "sister": 0.025553423794525788,
                "she": 0.0761118709786697,
                "her": 0.07224768225706711,
                "hers": 0.05228433658715302,
                "daughter": 0.019382166922557734,
                "male": 0.08512089128923325,
                "man": 0.07752571883094242,
                "boy": 0.03396126510372403,
                "brother": 0.030070032034284336,
                "he": 0.10583438336150637,
                "him": 0.09403512449772648,
                "his": 0.1058620066555313,
                "son": 0.027697936550083888,
            },
        }


        Since each run of RNSB may give a different result due to the random
        ``train_test_split`` and random initializations, RNSB can be requested to
        run many times and returns the average of all runs through the
        parameter ``n_iterations``.
        This makes it potentially more stable and robust to outlier runs.

        >>> RNSB().run_query(query, model, n_iterations=1000)
        {
            "query_name": "Female terms and Male Terms wrt Family and Careers",
            "result": 0.09649701346914233,
            "rnsb": 0.09649701346914233,
            "negative_sentiment_probabilities": {
                "female": 0.5618993210534083,
                "woman": 0.31188456697468364,
                "girl": 0.1968846981458747,
                "sister": 0.1666990161087616,
                "she": 0.4120315698794307,
                "her": 0.3956786125532543,
                "hers": 0.3031550094192968,
                "daughter": 0.13259627603249385,
                "male": 0.45579890258209677,
                "man": 0.4210218238530363,
                "boy": 0.2104231329680286,
                "brother": 0.18879207133574177,
                "he": 0.5473770682025214,
                "him": 0.4924664455586234,
                "his": 0.5479209229372095,
                "son": 0.1770764765027373,
            },
            "negative_sentiment_distribution": {
                "female": 0.10176190651838826,
                "woman": 0.056483371593163176,
                "girl": 0.035656498409823205,
                "sister": 0.030189767202716926,
                "she": 0.07462033949087124,
                "her": 0.07165876247453706,
                "hers": 0.05490241858857015,
                "daughter": 0.024013643264435475,
                "male": 0.08254675451251275,
                "man": 0.07624850551663455,
                "boy": 0.03810835568595322,
                "brother": 0.034190895761652934,
                "he": 0.09913187640146649,
                "him": 0.08918737310881396,
                "his": 0.09923037037111067,
                "son": 0.03206916109934998,
            },
        }

        If you want the embeddings to be normalized before calculating the metrics
        use the ``normalize=True`` before executing the query.

        >>> RNSB().run_query(query, model, normalize=True)
        {
            "query_name": "Female terms and Male Terms wrt Family and Careers",
            "result": 0.00957187793390364,
            "rnsb": 0.00957187793390364,
            "negative_sentiment_probabilities": {
                "female": 0.5078372178028085,
                "woman": 0.4334357574118245,
                "girl": 0.3764103216054252,
                "sister": 0.35256834229924383,
                "she": 0.4454357087596428,
                "her": 0.4390986149718311,
                "hers": 0.41329577968574494,
                "daughter": 0.33427930165282493,
                "male": 0.470250420503012,
                "man": 0.4577545228416623,
                "boy": 0.3698438702135818,
                "brother": 0.35380575403374315,
                "he": 0.49962008627445753,
                "him": 0.47052126448152776,
                "his": 0.49505591114138436,
                "son": 0.34192683607526553,
            },
            "negative_sentiment_distribution": {
                "female": 0.07511118533317328,
                "woman": 0.06410690741777263,
                "girl": 0.05567261405091151,
                "sister": 0.05214628855999085,
                "she": 0.06588174891831232,
                "her": 0.0649444670309599,
                "hers": 0.061128122983393235,
                "daughter": 0.04944126523085684,
                "male": 0.0695519454840733,
                "man": 0.06770375151119586,
                "boy": 0.054701409243182085,
                "brother": 0.05232930677698083,
                "he": 0.07389583823474007,
                "him": 0.06959200440758956,
                "his": 0.07322077821098569,
                "son": 0.050572366605882095,
            },
        }

        RNSB accepts more than 2 sets of target words.
        This example shows how to measure words representing different nationalities
        with respect to positive and negative words.

        Note this is one of the tests that was proposed in the RNSB paper. You can
        see the full paper replication in Previous Studies Replication.

        >>> import gensim.downloader as api
        >>>
        >>> from wefe.word_embedding_model import WordEmbeddingModel
        >>> from wefe.query import Query
        >>> from wefe.datasets import load_bingliu
        >>> from wefe.metrics import RNSB
        >>>
        >>> # Load the model
        >>> model = WordEmbeddingModel(
        ...        api.load('glove-wiki-gigaword-300'), 'Glove wiki'
        ...    )
        >>>
        >>> RNSB_words = [
        ...    ["swedish"],
        ...    ["irish"],
        ...    ["mexican"],
        ...    ["chinese"],
        ...    ["filipino"],
        ...    ["german"],
        ...    ["english"],
        ...    ["french"],
        ...    ["norwegian"],
        ...    ["american"],
        ...    ["indian"],
        ...    ["dutch"],
        ...    ["russian"],
        ...    ["scottish"],
        ...    ["italian"],
        ... ]
        >>>
        >>> bing_liu = load_bingliu()
        >>>
        >>> # Create the query
        >>> query = Query(
        ...     RNSB_words,
        ...     [bing_liu["positive_words"], bing_liu["negative_words"]],
        ...     attribute_sets_names=["Positive words", "Negative words"],
        ... )
        >>>
        >>> results = RNSB().run_query(
        ...     query,
        ...     model,
        ...     preprocessors=[{"lowercase": True}],
        ...     holdout=True,
        ...     print_model_evaluation=True,
        ...     n_iterations=500,
        ... )
        >>> results
        {
            "query_name": (
                "Target set 0, Target set 1, Target set 2, Target set 3, "
                "Target set 4, Target set 5, Target set 6, Target set 7, "
                "Target set 8, Target set 9, Target set 10, Target set 11, "
                "Target set 12, Target set 13 and Target set 14 "
                "wrt Positive words and Negative words"
            ),
            "result": 0.6313118439654091,
            "rnsb": 0.6313118439654091,
            "negative_sentiment_probabilities": {
                "swedish": 0.03865446713798508,
                "irish": 0.12266387930214015,
                "mexican": 0.5038405165657709,
                "chinese": 0.01913990969357335,
                "filipino": 0.08074140612507152,
                "german": 0.0498762435972975,
                "english": 0.058042779461913364,
                "french": 0.08030917713203162,
                "norwegian": 0.12177903128690087,
                "american": 0.22908203952254658,
                "indian": 0.7836948288757486,
                "dutch": 0.22748838866881654,
                "russian": 0.4877408793080844,
                "scottish": 0.027805085889223837,
                "italian": 0.007885923500742055,
            },
            "negative_sentiment_distribution": {
                "swedish": 0.01361674725376778,
                "irish": 0.04321060837966024,
                "mexican": 0.17748709213331876,
                "chinese": 0.006742385345191273,
                "filipino": 0.02844264587050847,
                "german": 0.017569824481278706,
                "english": 0.020446636995867174,
                "french": 0.028290385255119174,
                "norwegian": 0.04289890438595362,
                "american": 0.08069836330742804,
                "indian": 0.27607092269031136,
                "dutch": 0.08013697047258364,
                "russian": 0.17181569869170976,
                "scottish": 0.009794853090881381,
                "italian": 0.0027779616464207076,
            },
        }



        """
        # check the types of the provided arguments (only the defaults).
        self._check_input(
            query=query,
            model=model,
            lost_vocabulary_threshold=lost_vocabulary_threshold,
            warn_not_found_words=warn_not_found_words,
        )

        if n_iterations > 1 and random_state is not None:
            raise ValueError(
                "It is not possible to specify random_state together with n_iterations"
                " > 1 since all iterations would produce the same results."
            )

        # transform query word sets into embeddings
        embeddings = get_embeddings_from_query(
            model=model,
            query=query,
            lost_vocabulary_threshold=lost_vocabulary_threshold,
            preprocessors=preprocessors,
            strategy=strategy,
            normalize=normalize,
            warn_not_found_words=warn_not_found_words,
        )

        # if there is any/some set has less words than the allowed limit,
        # return the default value (nan)
        if embeddings is None:
            return {
                "query_name": query.query_name,
                "result": np.nan,
                "rnsb": np.nan,
                "negative_sentiment_probabilities": {},
                "negative_sentiment_distribution": {},
            }

        # get the targets and attribute sets transformed into embeddings.
        target_sets, attribute_sets = embeddings

        # get only the embeddings of the sets.
        target_embeddings = list(target_sets.values())
        attribute_embeddings = list(attribute_sets.values())

        # create the arrays that will contain the scores for each iteration
        calculated_divergences = []
        calculated_negative_sentiment_probabilities = []
        scores = []

        # calculate the scores for each iteration
        for i in range(n_iterations):
            try:
                if print_model_evaluation and (i > 0 and i < 2):
                    print(
                        "When n_iterations > 1, only the first evaluation is printed."
                    )
                    print_model_evaluation = False

                # train the logit with the train data.
                trained_classifier, score = self._train_classifier(
                    attribute_embeddings_dict=attribute_embeddings,
                    random_state=random_state,
                    estimator=estimator,
                    estimator_params=estimator_params,
                    holdout=holdout,
                    print_model_evaluation=print_model_evaluation,
                )

                scores.append(score)

                # get the scores
                divergence, negative_sentiment_probabilities = self._calc_rnsb(
                    target_embeddings, trained_classifier
                )

                calculated_divergences.append(divergence)
                calculated_negative_sentiment_probabilities.append(
                    negative_sentiment_probabilities
                )
            except Exception as e:
                logging.exception("RNSB Iteration omitted: " + str(e))

        # aggregate results
        divergence = np.mean(np.array(calculated_divergences))
        negative_sentiment_probabilities = dict(
            pd.DataFrame(calculated_negative_sentiment_probabilities).mean()
        )

        sum_of_prob = np.sum(list(negative_sentiment_probabilities.values()))
        negative_sentiment_distribution = {
            word: prob / sum_of_prob
            for word, prob in negative_sentiment_probabilities.items()
        }
        return {
            "query_name": query.query_name,
            "result": divergence,
            "rnsb": divergence,
            "negative_sentiment_probabilities": negative_sentiment_probabilities,
            "negative_sentiment_distribution": negative_sentiment_distribution,
        }
